# This file is part of cloud-init. See LICENSE file for license information.

import responses

from cloudinit import url_helper as uh
from cloudinit.sources.helpers import ec2
from tests.unittests import helpers


class TestEc2Util(helpers.ResponsesTestCase):
    VERSION = "latest"

    def test_userdata_fetch(self):
        self.responses.add(
            responses.GET,
            "http://169.254.169.254/%s/user-data" % (self.VERSION),
            body="stuff",
            status=200,
        )
        userdata = ec2.get_instance_userdata(self.VERSION)
        self.assertEqual("stuff", userdata.decode("utf-8"))

    def test_userdata_fetch_fail_not_found(self):
        self.responses.add(
            responses.GET,
            "http://169.254.169.254/%s/user-data" % (self.VERSION),
            status=404,
        )
        userdata = ec2.get_instance_userdata(self.VERSION, retries=0)
        self.assertEqual(b"", userdata)

    def test_userdata_fetch_fail_server_dead(self):
        self.responses.add(
            responses.GET,
            "http://169.254.169.254/%s/user-data" % (self.VERSION),
            status=500,
        )
        userdata = ec2.get_instance_userdata(self.VERSION, retries=0)
        self.assertEqual(b"", userdata)

    def test_userdata_fetch_fail_server_not_found(self):
        self.responses.add(
            responses.GET,
            "http://169.254.169.254/%s/user-data" % (self.VERSION),
            status=404,
        )
        userdata = ec2.get_instance_userdata(self.VERSION)
        self.assertEqual(b"", userdata)

    def test_metadata_fetch_no_keys(self):
        base_url = "http://169.254.169.254/%s/meta-data/" % (self.VERSION)
        self.responses.add(
            responses.GET,
            base_url,
            status=200,
            body="\n".join(["hostname", "instance-id", "ami-launch-index"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "hostname"),
            status=200,
            body="ec2.fake.host.name.com",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "instance-id"),
            status=200,
            body="123",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "ami-launch-index"),
            status=200,
            body="1",
        )
        md = ec2.get_instance_metadata(self.VERSION, retries=0)
        self.assertEqual(md["hostname"], "ec2.fake.host.name.com")
        self.assertEqual(md["instance-id"], "123")
        self.assertEqual(md["ami-launch-index"], "1")

    def test_metadata_fetch_key(self):
        base_url = "http://169.254.169.254/%s/meta-data/" % (self.VERSION)
        self.responses.add(
            responses.GET,
            base_url,
            status=200,
            body="\n".join(["hostname", "instance-id", "public-keys/"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "hostname"),
            status=200,
            body="ec2.fake.host.name.com",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "instance-id"),
            status=200,
            body="123",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "public-keys/"),
            status=200,
            body="0=my-public-key",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "public-keys/0/openssh-key"),
            status=200,
            body="ssh-rsa AAAA.....wZEf my-public-key",
        )
        md = ec2.get_instance_metadata(self.VERSION, retries=0, timeout=0.1)
        self.assertEqual(md["hostname"], "ec2.fake.host.name.com")
        self.assertEqual(md["instance-id"], "123")
        self.assertEqual(1, len(md["public-keys"]))

    def test_metadata_fetch_with_2_keys(self):
        base_url = "http://169.254.169.254/%s/meta-data/" % (self.VERSION)
        self.responses.add(
            responses.GET,
            base_url,
            status=200,
            body="\n".join(["hostname", "instance-id", "public-keys/"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "hostname"),
            status=200,
            body="ec2.fake.host.name.com",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "instance-id"),
            status=200,
            body="123",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "public-keys/"),
            status=200,
            body="\n".join(["0=my-public-key", "1=my-other-key"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "public-keys/0/openssh-key"),
            status=200,
            body="ssh-rsa AAAA.....wZEf my-public-key",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "public-keys/1/openssh-key"),
            status=200,
            body="ssh-rsa AAAA.....wZEf my-other-key",
        )
        md = ec2.get_instance_metadata(self.VERSION, retries=0, timeout=0.1)
        self.assertEqual(md["hostname"], "ec2.fake.host.name.com")
        self.assertEqual(md["instance-id"], "123")
        self.assertEqual(2, len(md["public-keys"]))

    def test_metadata_fetch_bdm(self):
        base_url = "http://169.254.169.254/%s/meta-data/" % (self.VERSION)
        self.responses.add(
            responses.GET,
            base_url,
            status=200,
            body="\n".join(
                ["hostname", "instance-id", "block-device-mapping/"]
            ),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "hostname"),
            status=200,
            body="ec2.fake.host.name.com",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "instance-id"),
            status=200,
            body="123",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "block-device-mapping/"),
            status=200,
            body="\n".join(["ami", "ephemeral0"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "block-device-mapping/ami"),
            status=200,
            body="sdb",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "block-device-mapping/ephemeral0"),
            status=200,
            body="sdc",
        )
        md = ec2.get_instance_metadata(self.VERSION, retries=0, timeout=0.1)
        self.assertEqual(md["hostname"], "ec2.fake.host.name.com")
        self.assertEqual(md["instance-id"], "123")
        bdm = md["block-device-mapping"]
        self.assertEqual(2, len(bdm))
        self.assertEqual(bdm["ami"], "sdb")
        self.assertEqual(bdm["ephemeral0"], "sdc")

    def test_metadata_no_security_credentials(self):
        base_url = "http://169.254.169.254/%s/meta-data/" % (self.VERSION)
        self.responses.add(
            responses.GET,
            base_url,
            status=200,
            body="\n".join(["instance-id", "iam/"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "instance-id"),
            status=200,
            body="i-0123451689abcdef0",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "iam/"),
            status=200,
            body="\n".join(["info/", "security-credentials/"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "iam/info/"),
            status=200,
            body="LastUpdated",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "iam/info/LastUpdated"),
            status=200,
            body="2016-10-27T17:29:39Z",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "iam/security-credentials/"),
            status=200,
            body="ReadOnly/",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "iam/security-credentials/ReadOnly/"),
            status=200,
            body="\n".join(["LastUpdated", "Expiration"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(
                base_url, "iam/security-credentials/ReadOnly/LastUpdated"
            ),
            status=200,
            body="2016-10-27T17:28:17Z",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(
                base_url, "iam/security-credentials/ReadOnly/Expiration"
            ),
            status=200,
            body="2016-10-28T00:00:34Z",
        )
        md = ec2.get_instance_metadata(self.VERSION, retries=0, timeout=0.1)
        self.assertEqual(md["instance-id"], "i-0123451689abcdef0")
        iam = md["iam"]
        self.assertEqual(1, len(iam))
        self.assertEqual(iam["info"]["LastUpdated"], "2016-10-27T17:29:39Z")
        self.assertNotIn("security-credentials", iam)

    def test_metadata_children_with_invalid_character(self):
        def _skip_tags(exception):
            if isinstance(exception, uh.UrlError) and exception.code == 404:
                if "meta-data/tags/" in exception.url:
                    print(exception.url)
                    return True
            return False

        base_url = "http://169.254.169.254/%s/meta-data/" % (self.VERSION)
        self.responses.add(
            responses.GET,
            base_url,
            status=200,
            body="\n".join(["tags/", "ami-launch-index"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "tags/"),
            status=200,
            body="\n".join(["test/invalid", "valid"]),
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "tags/valid"),
            status=200,
            body="OK",
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "tags/test/invalid"),
            status=404,
        )
        self.responses.add(
            responses.GET,
            uh.combine_url(base_url, "ami-launch-index"),
            status=200,
            body="1",
        )
        md = ec2.get_instance_metadata(
            self.VERSION,
            retries=0,
            timeout=0.1,
            retrieval_exception_ignore_cb=_skip_tags,
        )
        self.assertEqual(md["tags"]["valid"], "OK")
        self.assertEqual(md["tags"]["test/invalid"], "(skipped)")
        self.assertEqual(md["ami-launch-index"], "1")
        md = ec2.get_instance_metadata(self.VERSION, retries=0, timeout=0.1)
        self.assertEqual(len(md), 0)



# This file is part of cloud-init. See LICENSE file for license information.

import logging
import shutil
import tempfile

import pytest

from cloudinit import util
from cloudinit.config.cc_write_files_deferred import handle
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import (
    FilesystemMockingTestCase,
    skipUnlessJsonSchema,
)

LOG = logging.getLogger(__name__)


class TestWriteFilesDeferred(FilesystemMockingTestCase):

    with_logs = True

    def setUp(self):
        super(TestWriteFilesDeferred, self).setUp()
        self.tmp = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmp)

    def test_filtering_deferred_files(self):
        self.patchUtils(self.tmp)
        expected = "hello world\n"
        config = {
            "write_files": [
                {
                    "path": "/tmp/deferred.file",
                    "defer": True,
                    "content": expected,
                },
                {"path": "/tmp/not_deferred.file"},
            ]
        }
        cc = self.tmp_cloud("ubuntu")
        handle("cc_write_files_deferred", config, cc, [])
        self.assertEqual(util.load_text_file("/tmp/deferred.file"), expected)
        with self.assertRaises(FileNotFoundError):
            util.load_text_file("/tmp/not_deferred.file")


class TestWriteFilesDeferredSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # Allow undocumented keys client keys without error
            (
                {"write_files": [{"defer": "no"}]},
                "write_files.0.defer: 'no' is not of type 'boolean'",
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        with pytest.raises(SchemaValidationError, match=error_msg):
            validate_cloudconfig_schema(config, get_schema(), strict=True)



from typing import Optional

import cloudinit.net.netops as netops
from cloudinit import subp


class Iproute2(netops.NetOps):
    @staticmethod
    def link_up(interface: str, family: Optional[str] = None):
        subp.subp(
            ["ip"]
            + (["-family", family] if family else [])
            + ["link", "set", "dev", interface, "up"]
        )

    @staticmethod
    def link_down(interface: str, family: Optional[str] = None):
        subp.subp(
            ["ip"]
            + (["-family", family] if family else [])
            + ["link", "set", "dev", interface, "down"]
        )

    @staticmethod
    def add_route(
        interface: str,
        route: str,
        *,
        gateway: Optional[str] = None,
        source_address: Optional[str] = None,
    ):
        subp.subp(
            ["ip", "-4", "route", "add", route]
            + (["via", gateway] if gateway and gateway != "0.0.0.0" else [])
            + [
                "dev",
                interface,
            ]
            + (["src", source_address] if source_address else []),
        )

    @staticmethod
    def append_route(interface: str, address: str, gateway: str):
        subp.subp(
            ["ip", "-4", "route", "append", address]
            + (["via", gateway] if gateway and gateway != "0.0.0.0" else [])
            + ["dev", interface]
        )

    @staticmethod
    def del_route(
        interface: str,
        address: str,
        *,
        gateway: Optional[str] = None,
        source_address: Optional[str] = None,
    ):
        subp.subp(
            ["ip", "-4", "route", "del", address]
            + (["via", gateway] if gateway and gateway != "0.0.0.0" else [])
            + ["dev", interface]
            + (["src", source_address] if source_address else [])
        )

    @staticmethod
    def get_default_route() -> str:
        return subp.subp(
            ["ip", "route", "show", "0.0.0.0/0"],
        ).stdout

    @staticmethod
    def add_addr(interface: str, address: str, broadcast: str):
        subp.subp(
            [
                "ip",
                "-family",
                "inet",
                "addr",
                "add",
                address,
                "broadcast",
                broadcast,
                "dev",
                interface,
            ],
            update_env={"LANG": "C"},
        )

    @staticmethod
    def del_addr(interface: str, address: str):
        subp.subp(
            ["ip", "-family", "inet", "addr", "del", address, "dev", interface]
        )



# This file is part of cloud-init. See LICENSE file for license information.

import copy
import logging
from typing import List

from cloudinit import dmi, sources
from cloudinit.event import EventScope, EventType
from cloudinit.sources import DataSourceEc2 as EC2
from cloudinit.sources import DataSourceHostname

LOG = logging.getLogger(__name__)

ALIYUN_PRODUCT = "Alibaba Cloud ECS"


class DataSourceAliYun(EC2.DataSourceEc2):

    dsname = "AliYun"
    metadata_urls = ["http://100.100.100.200"]

    # The minimum supported metadata_version from the ec2 metadata apis
    min_metadata_version = "2016-01-01"
    extended_metadata_versions: List[str] = []

    # Aliyun metadata server security enhanced mode overwrite
    @property
    def imdsv2_token_put_header(self):
        return "X-aliyun-ecs-metadata-token"

    def __init__(self, sys_cfg, distro, paths):
        super(DataSourceAliYun, self).__init__(sys_cfg, distro, paths)
        self.default_update_events = copy.deepcopy(self.default_update_events)
        self.default_update_events[EventScope.NETWORK].add(EventType.BOOT)

    def get_hostname(self, fqdn=False, resolve_ip=False, metadata_only=False):
        hostname = self.metadata.get("hostname")
        is_default = False
        if hostname is None:
            hostname = "localhost.localdomain"
            is_default = True
        return DataSourceHostname(hostname, is_default)

    def get_public_ssh_keys(self):
        return parse_public_keys(self.metadata.get("public-keys", {}))

    def _get_cloud_name(self):
        if _is_aliyun():
            return EC2.CloudNames.ALIYUN
        else:
            return EC2.CloudNames.NO_EC2_METADATA


def _is_aliyun():
    return dmi.read_dmi_data("system-product-name") == ALIYUN_PRODUCT


def parse_public_keys(public_keys):
    keys = []
    for _key_id, key_body in public_keys.items():
        if isinstance(key_body, str):
            keys.append(key_body.strip())
        elif isinstance(key_body, list):
            keys.extend(key_body)
        elif isinstance(key_body, dict):
            key = key_body.get("openssh-key", [])
            if isinstance(key, str):
                keys.append(key.strip())
            elif isinstance(key, list):
                keys.extend(key)
    return keys


class DataSourceAliYunLocal(DataSourceAliYun):
    """Datasource run at init-local which sets up network to query metadata.

    In init-local, no network is available. This subclass sets up minimal
    networking with dhclient on a viable nic so that it can talk to the
    metadata service. If the metadata service provides network configuration
    then render the network configuration for that instance based on metadata.
    """

    perform_dhcp_setup = True


# Used to match classes to dependencies
datasources = [
    (DataSourceAliYunLocal, (sources.DEP_FILESYSTEM,)),  # Run at init-local
    (DataSourceAliYun, (sources.DEP_FILESYSTEM, sources.DEP_NETWORK)),
]

# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)



# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"Yum Add Repo: Add yum repository configuration to the system"

import io
import logging
import os
from configparser import ConfigParser
from textwrap import dedent

from cloudinit import util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.settings import PER_INSTANCE

MODULE_DESCRIPTION = """\
Add yum repository configuration to ``/etc/yum.repos.d``. Configuration files
are named based on the opaque dictionary key under the ``yum_repos`` they are
specified with. If a config file already exists with the same name as a config
entry, the config entry will be skipped.
"""

distros = [
    "almalinux",
    "centos",
    "cloudlinux",
    "eurolinux",
    "fedora",
    "mariner",
    "openeuler",
    "OpenCloudOS",
    "openmandriva",
    "photon",
    "rhel",
    "rocky",
    "TencentOS",
    "virtuozzo",
]

COPR_BASEURL = (
    "https://download.copr.fedorainfracloud.org/results/@cloud-init/"
    "cloud-init-dev/epel-8-$basearch/"
)
COPR_GPG_URL = (
    "https://download.copr.fedorainfracloud.org/results/@cloud-init/"
    "cloud-init-dev/pubkey.gpg"
)
EPEL_TESTING_BASEURL = (
    "https://download.copr.fedorainfracloud.org/results/@cloud-init/"
    "cloud-init-dev/pubkey.gpg"
)

meta: MetaSchema = {
    "id": "cc_yum_add_repo",
    "name": "Yum Add Repo",
    "title": "Add yum repository configuration to the system",
    "description": MODULE_DESCRIPTION,
    "distros": distros,
    "examples": [
        dedent(
            """\
            yum_repos:
              my_repo:
                baseurl: http://blah.org/pub/epel/testing/5/$basearch/
            yum_repo_dir: /store/custom/yum.repos.d
            """
        ),
        dedent(
            f"""\
            # Enable cloud-init upstream's daily testing repo for EPEL 8 to
            # install latest cloud-init from tip of `main` for testing.
            yum_repos:
              cloud-init-daily:
                name: Copr repo for cloud-init-dev owned by @cloud-init
                baseurl: {COPR_BASEURL}
                type: rpm-md
                skip_if_unavailable: true
                gpgcheck: true
                gpgkey: {COPR_GPG_URL}
                enabled_metadata: 1
            """
        ),
        dedent(
            f"""\
            # Add the file /etc/yum.repos.d/epel_testing.repo which can then
            # subsequently be used by yum for later operations.
            yum_repos:
            # The name of the repository
             epel-testing:
               baseurl: {EPEL_TESTING_BASEURL}
               enabled: false
               failovermethod: priority
               gpgcheck: true
               gpgkey: file:///etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL
               name: Extra Packages for Enterprise Linux 5 - Testing
            """
        ),
        dedent(
            """\
            # Any yum repo configuration can be passed directly into
            # the repository file created. See: man yum.conf for supported
            # config keys.
            #
            # Write /etc/yum.conf.d/my-package-stream.repo with gpgkey checks
            # on the repo data of the repository enabled.
            yum_repos:
              my package stream:
                baseurl: http://blah.org/pub/epel/testing/5/$basearch/
                mirrorlist: http://some-url-to-list-of-baseurls
                repo_gpgcheck: 1
                enable_gpgcheck: true
                gpgkey: https://url.to.ascii-armored-gpg-key
            """
        ),
    ],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": ["yum_repos"],
}

__doc__ = get_meta_doc(meta)
LOG = logging.getLogger(__name__)


def _canonicalize_id(repo_id: str) -> str:
    """Canonicalize repo id.

    The sole name convention for repo ids is to not contain namespaces,
    and typically the separator used is `-`. More info:
    https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/6/html/deployment_guide/sec-setting_repository_options

    :param repo_id: Repo id to convert.
    :return: Canonical repo id.
    """
    return repo_id.replace(" ", "-")


def _format_repo_value(val):
    if isinstance(val, (bool)):
        # Seems like yum prefers 1/0
        return str(int(val))
    if isinstance(val, (list, tuple)):
        # Can handle 'lists' in certain cases
        # See: https://linux.die.net/man/5/yum.conf
        return "\n".join([_format_repo_value(v) for v in val])
    if not isinstance(val, str):
        return str(val)
    return val


# TODO(harlowja): move to distro?
# See man yum.conf
def _format_repository_config(repo_id, repo_config):
    to_be = ConfigParser()
    to_be.add_section(repo_id)
    # Do basic translation of the items -> values
    for (k, v) in repo_config.items():
        # For now assume that people using this know
        # the format of yum and don't verify keys/values further
        to_be.set(repo_id, k, _format_repo_value(v))
    to_be_stream = io.StringIO()
    to_be.write(to_be_stream)
    to_be_stream.seek(0)
    lines = to_be_stream.readlines()
    lines.insert(0, "# Created by cloud-init on %s\n" % (util.time_rfc2822()))
    return "".join(lines)


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    repos = cfg.get("yum_repos")
    if not repos:
        LOG.debug(
            "Skipping module named %s, no 'yum_repos' configuration found",
            name,
        )
        return
    repo_base_path = util.get_cfg_option_str(
        cfg, "yum_repo_dir", "/etc/yum.repos.d/"
    )
    repo_locations = {}
    repo_configs = {}
    for (repo_id, repo_config) in repos.items():
        canon_repo_id = _canonicalize_id(repo_id)
        repo_fn_pth = os.path.join(repo_base_path, "%s.repo" % (canon_repo_id))
        if os.path.exists(repo_fn_pth):
            LOG.info(
                "Skipping repo %s, file %s already exists!",
                repo_id,
                repo_fn_pth,
            )
            continue
        elif canon_repo_id in repo_locations:
            LOG.info(
                "Skipping repo %s, file %s already pending!",
                repo_id,
                repo_fn_pth,
            )
            continue
        if not repo_config:
            repo_config = {}
        # Do some basic sanity checks/cleaning
        n_repo_config = {}
        for (k, v) in repo_config.items():
            k = k.lower().strip().replace("-", "_")
            if k:
                n_repo_config[k] = v
        repo_config = n_repo_config
        missing_required = 0
        for req_field in ["baseurl"]:
            if req_field not in repo_config:
                LOG.warning(
                    "Repository %s does not contain a %s"
                    " configuration 'required' entry",
                    repo_id,
                    req_field,
                )
                missing_required += 1
        if not missing_required:
            repo_configs[canon_repo_id] = repo_config
            repo_locations[canon_repo_id] = repo_fn_pth
        else:
            LOG.warning(
                "Repository %s is missing %s required fields, skipping!",
                repo_id,
                missing_required,
            )
    for (c_repo_id, path) in repo_locations.items():
        repo_blob = _format_repository_config(
            c_repo_id, repo_configs.get(c_repo_id)
        )
        util.write_file(path, repo_blob)



# Copyright (C) 2009-2010 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"""Byobu: Enable/disable byobu system wide and for default user."""

import logging

from cloudinit import subp, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.distros import ug_util
from cloudinit.settings import PER_INSTANCE

MODULE_DESCRIPTION = """\
This module controls whether byobu is enabled or disabled system wide and for
the default system user. If byobu is to be enabled, this module will ensure it
is installed. Likewise, if it is to be disabled, it will be removed if
installed.

Valid configuration options for this module are:

  - ``enable-system``: enable byobu system wide
  - ``enable-user``: enable byobu for the default user
  - ``disable-system``: disable byobu system wide
  - ``disable-user``: disable byobu for the default user
  - ``enable``: enable byobu both system wide and for default user
  - ``disable``: disable byobu for all users
  - ``user``: alias for ``enable-user``
  - ``system``: alias for ``enable-system``
"""
distros = ["ubuntu", "debian"]

LOG = logging.getLogger(__name__)

meta: MetaSchema = {
    "id": "cc_byobu",
    "name": "Byobu",
    "title": "Enable/disable byobu system wide and for default user",
    "description": MODULE_DESCRIPTION,
    "distros": distros,
    "frequency": PER_INSTANCE,
    "examples": [
        "byobu_by_default: enable-user",
        "byobu_by_default: disable-system",
    ],
    "activate_by_schema_keys": [],
}

__doc__ = get_meta_doc(meta)


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    if len(args) != 0:
        value = args[0]
    else:
        value = util.get_cfg_option_str(cfg, "byobu_by_default", "")

    if not value:
        LOG.debug("Skipping module named %s, no 'byobu' values found", name)
        return

    if value == "user" or value == "system":
        value = "enable-%s" % value

    valid = (
        "enable-user",
        "enable-system",
        "enable",
        "disable-user",
        "disable-system",
        "disable",
    )
    if value not in valid:
        LOG.warning("Unknown value %s for byobu_by_default", value)

    mod_user = value.endswith("-user")
    mod_sys = value.endswith("-system")
    if value.startswith("enable"):
        bl_inst = "install"
        dc_val = "byobu byobu/launch-by-default boolean true"
        mod_sys = True
    else:
        if value == "disable":
            mod_user = True
            mod_sys = True
        bl_inst = "uninstall"
        dc_val = "byobu byobu/launch-by-default boolean false"

    shcmd = ""
    if mod_user:
        (users, _groups) = ug_util.normalize_users_groups(cfg, cloud.distro)
        (user, _user_config) = ug_util.extract_default(users)
        if not user:
            LOG.warning(
                "No default byobu user provided, "
                "can not launch %s for the default user",
                bl_inst,
            )
        else:
            shcmd += ' sudo -Hu "%s" byobu-launcher-%s' % (user, bl_inst)
            shcmd += " || X=$(($X+1)); "
    if mod_sys:
        shcmd += 'echo "%s" | debconf-set-selections' % dc_val
        shcmd += " && dpkg-reconfigure byobu --frontend=noninteractive"
        shcmd += " || X=$(($X+1)); "

    if len(shcmd):
        cmd = ["/bin/sh", "-c", "%s %s %s" % ("X=0;", shcmd, "exit $X")]
        LOG.debug("Setting byobu to %s", value)
        subp.subp(cmd, capture=False)



"""Integration tests for various handlers."""

import re
from io import StringIO

import pytest

from cloudinit.cmd.devel.make_mime import create_mime_message
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM

PER_FREQ_TEMPLATE = """\
#!/bin/bash
touch /tmp/test_per_freq_{}
"""

PER_ALWAYS_FILE = StringIO(PER_FREQ_TEMPLATE.format("always"))
PER_INSTANCE_FILE = StringIO(PER_FREQ_TEMPLATE.format("instance"))
PER_ONCE_FILE = StringIO(PER_FREQ_TEMPLATE.format("once"))

FILES = [
    (PER_ALWAYS_FILE, "always.sh", "x-shellscript-per-boot"),
    (PER_INSTANCE_FILE, "instance.sh", "x-shellscript-per-instance"),
    (PER_ONCE_FILE, "once.sh", "x-shellscript-per-once"),
]

USER_DATA = create_mime_message(FILES)[0].as_string()


CLOUD_CONFIG_INVALID_YAML = """\
#cloud-config
ssh_import_id: [*]
"""
CLOUD_CONFIG_VALID = """\
#cloud-config
bootcmd: [echo yep]
"""

JINJA_INVALID = """\
## template:jinja
#cloud-config
runcmd:
{% for elem in (1, 2, 3) %}
- echo {{ elem }}
{% endor %}  # should be endfor
"""
JINJA_BAD_SCHEMA = """\
## template:jinja
#cloud-config
hostname: {{ range(1, 51) | random }}  # schema does not want int
"""
JINJA_VALID = """\
## template:jinja
#cloud-config
final_message:
    cloud-init the answer is: {{ 6 * 7 }}
"""

CLOUD_CONFIG_FILES = [
    (StringIO(CLOUD_CONFIG_INVALID_YAML), "cfg-invalid.yaml", "cloud-config"),
    (StringIO(CLOUD_CONFIG_VALID), "cfg-valid.yaml", "cloud-config"),
    (StringIO(JINJA_INVALID), "jinja-invalid.yaml", "jinja2"),
    (StringIO(JINJA_BAD_SCHEMA), "jinja-valid-invalid-schema.yaml", "jinja2"),
]
MIME_WITH_ERRORS = create_mime_message(CLOUD_CONFIG_FILES)[0].as_string()

CLOUD_CONFIG_ARCHIVE = """
#cloud-config-archive
- type: 'text/cloud-boothook'
  content: |
    #!/bin/sh
    echo "BOOTHOOK: $(date -R): this is called every boot." | tee /run/boothook.txt
- type: 'text/cloud-config'
  content: |
    bootcmd:
     - [sh, -c, 'echo "BOOTCMD: $(date -R): $INSTANCE_ID" | tee /run/bootcmd.txt']
"""  # noqa: E501


@pytest.mark.skipif(
    PLATFORM == "qemu", reason="QEMU only supports #cloud-config userdata"
)
@pytest.mark.ci
@pytest.mark.user_data(USER_DATA)
def test_per_freq(client: IntegrationInstance):
    # Sanity test for scripts folder
    cmd = "test -d /var/lib/cloud/scripts"
    assert client.execute(cmd).ok
    # Test per-boot
    cmd = "test -f /var/lib/cloud/scripts/per-boot/always.sh"
    assert client.execute(cmd).ok
    cmd = "test -f /tmp/test_per_freq_always"
    assert client.execute(cmd).ok
    # Test per-instance
    cmd = "test -f /var/lib/cloud/scripts/per-instance/instance.sh"
    assert client.execute(cmd).ok
    cmd = "test -f /tmp/test_per_freq_instance"
    assert client.execute(cmd).ok
    # Test per-once
    cmd = "test -f /var/lib/cloud/scripts/per-once/once.sh"
    assert client.execute(cmd).ok
    cmd = "test -f /tmp/test_per_freq_once"
    assert client.execute(cmd).ok
    client.restart()
    # Assert Always is run per boot
    cmd = "test -f /tmp/test_per_freq_always"
    assert client.execute(cmd).ok
    # Assert no per instance/once execution across reboot
    cmd = "test -f /tmp/test_per_freq_instance"
    assert client.execute(cmd).failed
    cmd = "test -f /tmp/test_per_freq_once"
    assert client.execute(cmd).failed


RE_EXPECTED_SCHEMA_STDERR = (
    r".*Error: Cloud config schema errors: format-l7.c1: Ignored invalid"
    r" user-data: cfg-invalid.yaml, hostname: \d+ is not of type 'string'"
)

RE_EXPECTED_SCHEMA_ANNOTATE = r"""#cloud-config

# from 2 files
# cfg-valid.yaml
# jinja-valid-invalid-schema.yaml

# Cloud-config part ignored SCHEMA_ERROR: cfg-invalid.yaml		# E1
---
bootcmd:
- echo yep
hostname: \d+		# E2
...

# Errors: -------------
# E1: Ignored invalid user-data: cfg-invalid.yaml
# E2: \d+ is not of type 'string'"""


@pytest.mark.skipif(
    PLATFORM == "qemu", reason="QEMU only supports #cloud-config userdata"
)
@pytest.mark.ci
@pytest.mark.user_data(MIME_WITH_ERRORS)
def test_mime_with_error_parts(client: IntegrationInstance):
    cmd = "cloud-init schema --system"
    result = client.execute(cmd)
    assert result.failed, f"Expected failure from {cmd}. Found: {result}"
    assert re.match(RE_EXPECTED_SCHEMA_STDERR, result.stderr)
    result = client.execute(f"{cmd} --annotate")
    assert re.findall(RE_EXPECTED_SCHEMA_ANNOTATE, result.stdout)
    log = client.read_from_file("/var/log/cloud-init.log")
    assert (
        "Failed at merging in cloud config part from cfg-invalid.yaml" in log
    )


@pytest.mark.ci
@pytest.mark.user_data(CLOUD_CONFIG_ARCHIVE)
def test_cloud_config_archive_boot_hook_logging(client: IntegrationInstance):
    """
    boot-hook and bootcmd scripts run per boot and log to cloud-init-outout.log
    """
    cmd = "cloud-init schema --system"
    assert "Valid schema user-data" in client.execute(cmd).stdout
    client.restart()
    log = client.read_from_file("/var/log/cloud-init-output.log")
    assert 2 == len(re.findall("BOOTHOOK:.*", log))
    assert 2 == len(re.findall("BOOTCMD:.*", log))



import pathlib
import re

import enchant


class WordListFilter(enchant.tokenize.Filter):
    word_list = "spelling_word_list.txt"
    regex_list = "spelling_regex_list.txt"

    def __init__(self, *args, **kwargs):
        """Use two files for ignoring correctly spelled words

        - spelling_word_list.txt: a list of exact matches to ignore
        - spelling_regex_list.txt: a list of regular expressions to ignore

        Splits tokens on "/" and "-".
        """
        super().__init__(*args, *kwargs)
        directory = pathlib.Path(__file__).parent
        with open(directory.joinpath(self.word_list)) as f:
            lines = f.read().splitlines()
            self._validate_lines(lines)
            self.word_set = set(lines)
            print(f"Loaded {self.word_list}: {lines})")
        with open(directory.joinpath(self.regex_list)) as f:
            regex_lines = f.read().splitlines()
            self.regex_set = set(regex_lines)
            print(f"Loaded {self.regex_list}: {regex_lines}")

    def _validate_lines(self, lines):
        """Assert that the word_list file is legible and orderly"""
        for line in lines:
            if line != line.lower():
                raise Exception(
                    f"Uppercase characters in {self.word_list} detected. "
                    "Please use lowercase characters for legibility."
                )
        if lines != sorted(lines):
            first_missordered = next_item = previous_item = None
            for item_a, item_b in zip(lines, sorted(lines)):
                if first_missordered:
                    next_item = item_a
                    break
                elif item_a != item_b:
                    first_missordered = item_a
                else:
                    previous_item = item_a
            unordered = (
                f"[..., {previous_item}, {first_missordered}, "
                f"{next_item}, ...]"
            )
            raise Exception(
                f"Unsorted {self.word_list} detected. "
                f"Please sort for legibility. Unordered list: {unordered}"
            )

    def _in_word_list(self, word):
        """Lowercase match the set of words in spelling_word_list.txt"""
        return word.lower() in self.word_set

    def _in_word_regex(self, word):
        """Regex match the expressions in spelling_regex_list.txt"""
        for regex in self.regex_set:
            out = re.search(regex, word)
            if out:
                return True

    def _skip(self, word):
        """Skip words and regex expressions in the allowlist files"""
        return self._in_word_list(word) or self._in_word_regex(word)

    def _split(self, word):
        """split words into sub-tokens on - and /"""
        if "-" in word or "/" in word:
            for i, token in enumerate(re.split("-|/", word)):
                if self._skip(token):
                    continue
                yield token, i



import logging

import pytest

from tests.integration_tests.clouds import IntegrationCloud
from tests.integration_tests.conftest import get_validated_source
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.util import lxd_has_nocloud, wait_for_cloud_init

log = logging.getLogger("integration_testing")


def restart_cloud_init(c):
    client = c
    client.instance.shutdown(wait=False)
    try:
        client.instance.wait_for_state("STOPPED", num_retries=20)
    except RuntimeError as e:
        log.warning(
            "Retrying shutdown due to timeout on initial shutdown request %s",
            str(e),
        )
        client.instance.shutdown()

    client.instance.execute_via_ssh = False
    client.instance.start()
    client.execute("cloud-init status --wait")


def override_kernel_cmdline(ds_str: str, c: IntegrationInstance):
    """
    Configure grub's kernel command line to tell cloud-init to use OpenStack
    - even though LXD should naturally be detected.

    This runs on LXD, but forces cloud-init to attempt to run OpenStack.
    This will inevitably fail on LXD, but we only care that it tried - on
    Ironic, for example, it will succeed.
    """
    client = c

    # The final output in /etc/default/grub should be:
    #
    # GRUB_CMDLINE_LINUX="'ds=nocloud;s=http://my-url/'"
    #
    # That ensures that the kernel commandline passed into
    # /boot/efi/EFI/ubuntu/grub.cfg will be properly single-quoted
    #
    # Example:
    #
    # linux /boot/vmlinuz-5.15.0-1030-kvm ro 'ds=nocloud;s=http://my-url/'
    #
    # Not doing this will result in a semicolon-delimited ds argument
    # terminating the kernel arguments prematurely.
    client.execute('printf "GRUB_CMDLINE_LINUX=\\"" >> /etc/default/grub')
    client.execute('printf "\'" >> /etc/default/grub')
    client.execute(f"printf '{ds_str}' >> /etc/default/grub")
    client.execute('printf "\'\\"" >> /etc/default/grub')

    # We should probably include non-systemd distros at some point. This should
    # most likely be as simple as updating the output path for grub-mkconfig
    client.execute("grub-mkconfig -o /boot/efi/EFI/ubuntu/grub.cfg")
    client.execute("cloud-init clean --logs")
    restart_cloud_init(client)


@pytest.mark.skipif(PLATFORM != "lxd_vm", reason="Modifies grub config")
@pytest.mark.lxd_use_exec
@pytest.mark.parametrize(
    "ds_str, configured, cmdline_configured",
    (
        (
            "ds=nocloud;s=http://my-url/;h=hostname",
            "DataSourceNoCloud [seed=None][dsmode=net]",
            True,
        ),
        ("ci.ds=openstack", "DataSourceOpenStack", True),
        ("bonding.max_bonds=0", "lxd_or_nocloud", False),
    ),
)
def test_lxd_datasource_kernel_override(
    ds_str, configured, cmdline_configured, client: IntegrationInstance
):
    """This test is twofold: it tests kernel commandline override, which also
    validates OpenStack Ironic requirements. OpenStack Ironic does not
    advertise itself to cloud-init via any of the conventional methods: DMI,
    etc.

    On systemd, ds-identify is able to grok kernel commandline, however to
    support cloud-init kernel command line parsing on non-systemd, parsing
    kernel commandline in Python code is required.
    """

    if configured == "lxd_or_nocloud":
        configured = (
            "DataSourceNoCloud" if lxd_has_nocloud(client) else "DataSourceLXD"
        )
    override_kernel_cmdline(ds_str, client)
    if cmdline_configured:
        assert (
            "Machine is configured by the kernel commandline to run on single "
            f"datasource {configured}"
        ) in client.execute("cat /var/log/cloud-init.log")
    else:
        # verify that no plat
        log = client.execute("cat /var/log/cloud-init.log")
        assert (f"Detected platform: {configured}") in log
        assert (
            "Machine is configured by the kernel "
            "commandline to run on single "
        ) not in log


GH_REPO_PATH = "https://raw.githubusercontent.com/canonical/cloud-init/main/"


@pytest.mark.skipif(PLATFORM != "lxd_vm", reason="Modifies grub config")
@pytest.mark.lxd_use_exec
@pytest.mark.parametrize(
    "ds_str",
    (f"ds=nocloud-net;s={GH_REPO_PATH}tests/data/kernel_cmdline_match/",),
)
def test_lxd_datasource_kernel_override_nocloud_net(
    ds_str, session_cloud: IntegrationCloud
):
    """NoCloud requirements vary slightly from other datasources with parsing
    nocloud-net due to historical reasons. Do to this variation, this is
    implemented in NoCloud's ds_detect and therefore has a different log
    message.
    """
    _ds_name, _, seed_url = ds_str.partition(";")
    _key, _, url_val = seed_url.partition("=")
    source = get_validated_source(session_cloud)
    with session_cloud.launch(
        wait=False,  # to prevent cloud-init status --wait
        launch_kwargs={
            # On Jammy and above, we detect the LXD datasource using a
            # socket available to the container. This prevents the socket
            # from being exposed in the container, so LXD will not be detected.
            # This allows us to wait for detection in 'init' stage with
            # DataSourceNoCloudNet.
            "config_dict": {"security.devlxd": False},
        },
    ) as client:
        # We know this will be an LXD instance due to our pytest mark
        client.instance.execute_via_ssh = False  # pyright: ignore
        assert wait_for_cloud_init(client, num_retries=60).ok
        if source.installs_new_version():
            client.install_new_cloud_init(source, clean=False)
        override_kernel_cmdline(ds_str, client)

        logs = client.execute("cat /var/log/cloud-init.log")
        assert (
            "nocloud"
            == client.execute("cloud-init query platform").stdout.strip()
        )
        assert url_val in client.execute("cloud-init query subplatform").stdout
        assert (
            "Detected platform: DataSourceNoCloudNet [seed=None]"
            "[dsmode=net]. Checking for active instance data"
        ) in logs


@pytest.mark.skipif(PLATFORM != "lxd_vm", reason="Modifies grub config")
@pytest.mark.lxd_use_exec
def test_lxd_disable_cloud_init_cmdline(client: IntegrationInstance):
    """Verify cloud-init disablement via kernel commandline works."""

    override_kernel_cmdline("cloud-init=disabled", client)
    assert "Active: inactive (dead)" in client.execute(
        "systemctl status cloud-init"
    )


@pytest.mark.lxd_use_exec
def test_lxd_disable_cloud_init_file(client: IntegrationInstance):
    """Verify cloud-init disablement via file works."""

    client.execute("touch /etc/cloud/cloud-init.disabled")
    client.execute("cloud-init --clean")
    restart_cloud_init(client)
    assert "Active: inactive (dead)" in client.execute(
        "systemctl status cloud-init"
    )


@pytest.mark.lxd_use_exec
def test_lxd_disable_cloud_init_env(client: IntegrationInstance):
    """Verify cloud-init disablement via environment variable works."""
    env = """DefaultEnvironment=KERNEL_CMDLINE=cloud-init=disabled"""

    client.execute(f'echo "{env}" >> /etc/systemd/system.conf')

    client.execute("cloud-init --clean")
    restart_cloud_init(client)
    assert "Active: inactive (dead)" in client.execute(
        "systemctl status cloud-init"
    )



# This file is part of cloud-init. See LICENSE file for license information.
import re

import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.releases import IS_UBUNTU

USER_DATA = """\
#cloud-config
write_files:
- path: /etc/network/interfaces.d/50-cloud-init.cfg
  content: cloud-init test was here
  permissions: '0644'
  owner: root:root
- path: /etc/cloud/clean.d/runme.sh
  encoding: b64
  content: IyEvYmluL3NoCmVjaG8gIi9ldGMvY2xvdWQvY2xlYW4uZC9ydW5tZS5zaCBSQU4iCg==
  permissions: '0755'
  owner: root:root
- path: /etc/systemd/network/10-cloud-init-eth0.network
  content: cloud-init test was here
  permissions: '0644'
  owner: root:root
- path: /etc/cloud/clean.d/dontrunme.sh
  content: '#!/bin/sh\necho DID NOT RUN BECAUSE NO EXEC PERMS'
  permissions: '0644'
  owner: root:root
packages:
- logrotate
"""


@pytest.mark.user_data(USER_DATA)
class TestCleanCommand:
    @pytest.mark.skipif(
        not IS_UBUNTU, reason="Hasn't been tested on other distros"
    )
    def test_clean_rotated_logs(self, client: IntegrationInstance):
        """Clean with log params alters expected files without error"""
        assert client.execute("cloud-init status --wait --long").ok
        assert client.execute("logrotate /etc/logrotate.d/cloud-init").ok
        log_paths = (
            "/var/log/cloud-init.log",
            "/var/log/cloud-init.log.1.gz",
            "/var/log/cloud-init-output.log",
            "/var/log/cloud-init-output.log.1.gz",
        )

        assert client.execute("cloud-init clean --logs").ok
        for path in log_paths:
            assert client.execute(
                f"test -f {path}"
            ).failed, f"Unexpected file found {path}"

    def test_clean_by_param(self, client: IntegrationInstance):
        """Clean with various params alters expected files without error"""
        assert client.execute("cloud-init status --wait --long").ok
        result = client.execute("cloud-init clean")
        assert (
            result.ok
        ), "non-zero exit on cloud-init clean runparts of /etc/cloud/clean.d"
        # Log files are not removed without --logs
        log_paths = (
            "/var/log/cloud-init.log",
            "/var/log/cloud-init-output.log",
        )
        net_cfg_paths = (
            "/etc/network/interfaces.d/50-cloud-init.cfg",
            "/etc/netplan/50-cloud-init.yaml",
            "/etc/systemd/network/10-cloud-init-eth0.network",
        )
        for path in log_paths + net_cfg_paths:
            assert client.execute(
                f"test -f {path}"
            ).ok, f"Missing expected file {path}"
        # /etc/cloud/clean.d runparts scripts are run if executable
        assert result.stdout == "/etc/cloud/clean.d/runme.sh RAN"

        # Log files removed with --logs
        assert client.execute("cloud-init clean --logs").ok
        for path in log_paths:
            assert client.execute(
                f"test -f {path}"
            ).failed, f"Unexpected file found {path}"
        for path in net_cfg_paths:
            assert client.execute(
                f"test -f {path}"
            ).ok, f"Missing expected file {path}"

        prev_machine_id = client.read_from_file("/etc/machine-id")
        assert re.match(
            r"^[a-f0-9]{32}$", prev_machine_id
        ), f"Unexpected machine-id format {prev_machine_id}"

        # --machine-id sets /etc/machine-id
        assert client.execute("cloud-init clean --machine-id").ok
        machine_id = client.read_from_file("/etc/machine-id")
        assert machine_id != prev_machine_id
        assert "uninitialized" == machine_id

        # --configs remove network scope
        assert client.execute("cloud-init clean --configs network").ok

        for path in log_paths + net_cfg_paths:
            assert client.execute(
                f"test -f {path}"
            ).failed, f"Unexpected file found {path}"



# This file is part of cloud-init. See LICENSE file for license information.

import copy
import json
import threading
from unittest import mock

import pytest
import requests
import responses

from cloudinit import helpers
from cloudinit.net import activators
from cloudinit.sources import DataSourceEc2 as ec2
from tests.unittests import helpers as test_helpers
from tests.unittests.util import MockDistro

DYNAMIC_METADATA = {
    "instance-identity": {
        "document": json.dumps(
            {
                "devpayProductCodes": None,
                "marketplaceProductCodes": ["1abc2defghijklm3nopqrs4tu"],
                "availabilityZone": "us-west-2b",
                "privateIp": "10.158.112.84",
                "version": "2017-09-30",
                "instanceId": "my-identity-id",
                "billingProducts": None,
                "instanceType": "t2.micro",
                "accountId": "123456789012",
                "imageId": "ami-5fb8c835",
                "pendingTime": "2016-11-19T16:32:11Z",
                "architecture": "x86_64",
                "kernelId": None,
                "ramdiskId": None,
                "region": "us-west-2",
            }
        )
    }
}


# collected from api version 2016-09-02/ with
# python3 -c 'import json
# from cloudinit.sources.helpers.ec2 import get_instance_metadata as gm
# print(json.dumps(gm("2016-09-02"), indent=1, sort_keys=True))'
# Note that the MAC addresses have been modified to sort in the opposite order
# to the device-number attribute, to test LP: #1876312
DEFAULT_METADATA = {
    "ami-id": "ami-8b92b4ee",
    "ami-launch-index": "0",
    "ami-manifest-path": "(unknown)",
    "block-device-mapping": {"ami": "/dev/sda1", "root": "/dev/sda1"},
    "hostname": "ip-172-31-31-158.us-east-2.compute.internal",
    "instance-action": "none",
    "instance-id": "i-0a33f80f09c96477f",
    "instance-type": "t2.small",
    "local-hostname": "ip-172-3-3-15.us-east-2.compute.internal",
    "local-ipv4": "172.3.3.15",
    "mac": "06:17:04:d7:26:09",
    "metrics": {"vhostmd": '<?xml version="1.0" encoding="UTF-8"?>'},
    "network": {
        "interfaces": {
            "macs": {
                "06:17:04:d7:26:09": {
                    "device-number": "0",
                    "interface-id": "eni-e44ef49e",
                    "ipv4-associations": {"13.59.77.202": "172.3.3.15"},
                    "ipv6s": "2600:1f16:aeb:b20b:9d87:a4af:5cc9:73dc",
                    "local-hostname": (
                        "ip-172-3-3-15.us-east-2.compute.internal"
                    ),
                    "local-ipv4s": "172.3.3.15",
                    "mac": "06:17:04:d7:26:09",
                    "owner-id": "950047163771",
                    "public-hostname": (
                        "ec2-13-59-77-202.us-east-2.compute.amazonaws.com"
                    ),
                    "public-ipv4s": "13.59.77.202",
                    "security-group-ids": "sg-5a61d333",
                    "security-groups": "wide-open",
                    "subnet-id": "subnet-20b8565b",
                    "subnet-ipv4-cidr-block": "172.31.16.0/20",
                    "subnet-ipv6-cidr-blocks": "2600:1f16:aeb:b20b::/64",
                    "vpc-id": "vpc-87e72bee",
                    "vpc-ipv4-cidr-block": "172.31.0.0/16",
                    "vpc-ipv4-cidr-blocks": "172.31.0.0/16",
                    "vpc-ipv6-cidr-blocks": "2600:1f16:aeb:b200::/56",
                },
                "06:17:04:d7:26:08": {
                    "device-number": "1",  # Only IPv4 local config
                    "interface-id": "eni-e44ef49f",
                    "ipv4-associations": {"": "172.3.3.16"},
                    "ipv6s": "",  # No IPv6 config
                    "local-hostname": (
                        "ip-172-3-3-16.us-east-2.compute.internal"
                    ),
                    "local-ipv4s": "172.3.3.16",
                    "mac": "06:17:04:d7:26:08",
                    "owner-id": "950047163771",
                    "public-hostname": (
                        "ec2-172-3-3-16.us-east-2.compute.amazonaws.com"
                    ),
                    "public-ipv4s": "",  # No public ipv4 config
                    "security-group-ids": "sg-5a61d333",
                    "security-groups": "wide-open",
                    "subnet-id": "subnet-20b8565b",
                    "subnet-ipv4-cidr-block": "172.31.16.0/20",
                    "subnet-ipv6-cidr-blocks": "",
                    "vpc-id": "vpc-87e72bee",
                    "vpc-ipv4-cidr-block": "172.31.0.0/16",
                    "vpc-ipv4-cidr-blocks": "172.31.0.0/16",
                    "vpc-ipv6-cidr-blocks": "",
                },
            }
        }
    },
    "placement": {"availability-zone": "us-east-2b"},
    "profile": "default-hvm",
    "public-hostname": "ec2-13-59-77-202.us-east-2.compute.amazonaws.com",
    "public-ipv4": "13.59.77.202",
    "public-keys": {"brickies": ["ssh-rsa AAAAB3Nz....w== brickies"]},
    "reservation-id": "r-01efbc9996bac1bd6",
    "security-groups": "my-wide-open",
    "services": {"domain": "amazonaws.com", "partition": "aws"},
}

# collected from api version 2018-09-24/ with
# python3 -c 'import json
# from cloudinit.sources.helpers.ec2 import get_instance_metadata as gm
# print(json.dumps(gm("2018-09-24"), indent=1, sort_keys=True))'

NIC1_MD_IPV4_IPV6_MULTI_IP = {
    "device-number": "0",
    "interface-id": "eni-0d6335689899ce9cc",
    "ipv4-associations": {"18.218.219.181": "172.31.44.13"},
    "ipv6s": [
        "2600:1f16:292:100:c187:593c:4349:136",
        "2600:1f16:292:100:f153:12a3:c37c:11f9",
        "2600:1f16:292:100:f152:2222:3333:4444",
    ],
    "local-hostname": "ip-172-31-44-13.us-east-2.compute.internal",
    "local-ipv4s": ["172.31.44.13", "172.31.45.70"],
    "mac": "0a:07:84:3d:6e:38",
    "owner-id": "329910648901",
    "public-hostname": "ec2-18-218-219-181.us-east-2.compute.amazonaws.com",
    "public-ipv4s": "18.218.219.181",
    "security-group-ids": "sg-0c387755222ba8d2e",
    "security-groups": "launch-wizard-4",
    "subnet-id": "subnet-9d7ba0d1",
    "subnet-ipv4-cidr-block": "172.31.32.0/20",
    "subnet_ipv6_cidr_blocks": "2600:1f16:292:100::/64",
    "vpc-id": "vpc-a07f62c8",
    "vpc-ipv4-cidr-block": "172.31.0.0/16",
    "vpc-ipv4-cidr-blocks": "172.31.0.0/16",
    "vpc_ipv6_cidr_blocks": "2600:1f16:292:100::/56",
}

NIC2_MD = {
    "device-number": "1",
    "interface-id": "eni-043cdce36ded5e79f",
    "local-hostname": "ip-172-31-47-221.us-east-2.compute.internal",
    "local-ipv4s": "172.31.47.221",
    "mac": "0a:75:69:92:e2:16",
    "owner-id": "329910648901",
    "security-group-ids": "sg-0d68fef37d8cc9b77",
    "security-groups": "launch-wizard-17",
    "subnet-id": "subnet-9d7ba0d1",
    "subnet-ipv4-cidr-block": "172.31.32.0/20",
    "vpc-id": "vpc-a07f62c8",
    "vpc-ipv4-cidr-block": "172.31.0.0/16",
    "vpc-ipv4-cidr-blocks": "172.31.0.0/16",
}

NIC2_MD_IPV4_IPV6_MULTI_IP = {
    "device-number": "1",
    "interface-id": "eni-043cdce36ded5e79f",
    "ipv6s": [
        "2600:1f16:292:100:c187:593c:4349:136",
        "2600:1f16:292:100:f153:12a3:c37c:11f9",
    ],
    "local-hostname": "ip-172-31-47-221.us-east-2.compute.internal",
    "local-ipv4s": "172.31.47.221",
    "mac": "0a:75:69:92:e2:16",
    "owner-id": "329910648901",
    "security-group-ids": "sg-0d68fef37d8cc9b77",
    "security-groups": "launch-wizard-17",
    "subnet-id": "subnet-9d7ba0d1",
    "subnet-ipv4-cidr-block": "172.31.32.0/20",
    "subnet-ipv6-cidr-blocks": "2600:1f16:292:100::/64",
    "vpc-id": "vpc-a07f62c8",
    "vpc-ipv4-cidr-block": "172.31.0.0/16",
    "vpc-ipv4-cidr-blocks": "172.31.0.0/16",
    "vpc-ipv6-cidr-blocks": "2600:1f16:292:100::/56",
}

SECONDARY_IP_METADATA_2018_09_24 = {
    "ami-id": "ami-0986c2ac728528ac2",
    "ami-launch-index": "0",
    "ami-manifest-path": "(unknown)",
    "block-device-mapping": {"ami": "/dev/sda1", "root": "/dev/sda1"},
    "events": {"maintenance": {"history": "[]", "scheduled": "[]"}},
    "hostname": "ip-172-31-44-13.us-east-2.compute.internal",
    "identity-credentials": {
        "ec2": {
            "info": {
                "AccountId": "329910648901",
                "Code": "Success",
                "LastUpdated": "2019-07-06T14:22:56Z",
            }
        }
    },
    "instance-action": "none",
    "instance-id": "i-069e01e8cc43732f8",
    "instance-type": "t2.micro",
    "local-hostname": "ip-172-31-44-13.us-east-2.compute.internal",
    "local-ipv4": "172.31.44.13",
    "mac": "0a:07:84:3d:6e:38",
    "metrics": {"vhostmd": '<?xml version="1.0" encoding="UTF-8"?>'},
    "network": {
        "interfaces": {
            "macs": {
                "0a:07:84:3d:6e:38": NIC1_MD_IPV4_IPV6_MULTI_IP,
            }
        }
    },
    "placement": {"availability-zone": "us-east-2c"},
    "profile": "default-hvm",
    "public-hostname": "ec2-18-218-219-181.us-east-2.compute.amazonaws.com",
    "public-ipv4": "18.218.219.181",
    "public-keys": {"yourkeyname,e": ["ssh-rsa AAAAW...DZ yourkeyname"]},
    "reservation-id": "r-09b4917135cdd33be",
    "security-groups": "launch-wizard-4",
    "services": {"domain": "amazonaws.com", "partition": "aws"},
}

M_PATH = "cloudinit.sources.DataSourceEc2."
M_PATH_NET = "cloudinit.sources.DataSourceEc2.net."

TAGS_METADATA_2021_03_23: dict = {
    **DEFAULT_METADATA,
    "tags": {
        "instance": {
            "Environment": "production",
            "Application": "test",
            "TagWithoutValue": "",
        }
    },
}


@pytest.fixture(autouse=True)
def disable_is_resolvable():
    with mock.patch("cloudinit.sources.DataSourceEc2.util.is_resolvable"):
        yield


def _register_ssh_keys(rfunc, base_url, keys_data):
    """handle ssh key inconsistencies.

    public-keys in the ec2 metadata is inconsistently formated compared
    to other entries.
    Given keys_data of {name1: pubkey1, name2: pubkey2}

    This registers the following urls:
       base_url                 0={name1}\n1={name2} # (for each name)
       base_url/                0={name1}\n1={name2} # (for each name)
       base_url/0               openssh-key
       base_url/0/              openssh-key
       base_url/0/openssh-key   {pubkey1}
       base_url/0/openssh-key/  {pubkey1}
       ...
    """

    base_url = base_url.rstrip("/")
    odd_index = "\n".join(
        ["{0}={1}".format(n, name) for n, name in enumerate(sorted(keys_data))]
    )

    rfunc(base_url, odd_index)
    rfunc(base_url + "/", odd_index)

    for n, name in enumerate(sorted(keys_data)):
        val = keys_data[name]
        if isinstance(val, list):
            val = "\n".join(val)
        burl = base_url + "/%s" % n
        rfunc(burl, "openssh-key")
        rfunc(burl + "/", "openssh-key")
        rfunc(burl + "/%s/openssh-key" % name, val)
        rfunc(burl + "/%s/openssh-key/" % name, val)


def register_mock_metaserver(base_url, data, responses_mock=None):
    """Register with responses a ec2 metadata like service serving 'data'.

    If given a dictionary, it will populate urls under base_url for
    that dictionary.  For example, input of
       {"instance-id": "i-abc", "mac": "00:16:3e:00:00:00"}
    populates
       base_url  with 'instance-id\nmac'
       base_url/ with 'instance-id\nmac'
       base_url/instance-id with i-abc
       base_url/mac with 00:16:3e:00:00:00
    In the index, references to lists or dictionaries have a trailing /.
    """
    responses_mock = responses_mock or responses

    def register_helper(register, base_url, body):
        if not isinstance(base_url, str):
            register(base_url, body)
            return
        base_url = base_url.rstrip("/")
        if isinstance(body, str):
            register(base_url, body)
        elif isinstance(body, list):
            register(base_url, "\n".join(body) + "\n")
            register(base_url + "/", "\n".join(body) + "\n")
        elif isinstance(body, dict):
            vals = []
            for k, v in body.items():
                if k == "public-keys":
                    _register_ssh_keys(register, base_url + "/public-keys/", v)
                    continue
                suffix = k.rstrip("/")
                if not isinstance(v, (str, list)):
                    suffix += "/"
                vals.append(suffix)
                url = base_url + "/" + suffix
                register_helper(register, url, v)
            register(base_url, "\n".join(vals) + "\n")
            register(base_url + "/", "\n".join(vals) + "\n")
        elif body is None:
            register(base_url, "not found", status=404)

    def myreg(*argc, **kwargs):
        url, body = argc
        method = responses.PUT if "latest/api/token" in url else responses.GET
        status = kwargs.get("status", 200)
        return responses_mock.add(method, url, body, status=status)

    register_helper(myreg, base_url, data)


class TestEc2(test_helpers.ResponsesTestCase):
    with_logs = True
    maxDiff = None

    valid_platform_data = {
        "uuid": "ec212f79-87d1-2f1d-588f-d86dc0fd5412",
        "uuid_source": "dmi",
        "serial": "ec212f79-87d1-2f1d-588f-d86dc0fd5412",
    }

    def setUp(self):
        super(TestEc2, self).setUp()
        self.datasource = ec2.DataSourceEc2
        self.metadata_addr = self.datasource.metadata_urls[0]
        self.tmp = self.tmp_dir()

    def data_url(self, version, data_item="meta-data"):
        """Return a metadata url based on the version provided."""
        return "/".join([self.metadata_addr, version, data_item])

    def _patch_add_cleanup(self, mpath, *args, **kwargs):
        p = mock.patch(mpath, *args, **kwargs)
        p.start()
        self.addCleanup(p.stop)

    def _setup_ds(
        self, sys_cfg, platform_data, md, md_version=None, distro=None
    ):
        self.uris = []
        distro = distro or mock.MagicMock()
        distro.get_tmp_exec_path = self.tmp_dir
        paths = helpers.Paths({"run_dir": self.tmp})
        if sys_cfg is None:
            sys_cfg = {}
        ds = self.datasource(sys_cfg=sys_cfg, distro=distro, paths=paths)
        event = threading.Event()
        p = mock.patch("time.sleep", event.wait)
        p.start()

        def _mock_sleep():
            event.set()
            p.stop()

        self.addCleanup(_mock_sleep)
        if not md_version:
            md_version = ds.min_metadata_version
        if platform_data is not None:
            self._patch_add_cleanup(
                "cloudinit.sources.DataSourceEc2._collect_platform_data",
                return_value=platform_data,
            )

        if md:
            all_versions = [
                ds.min_metadata_version
            ] + ds.extended_metadata_versions
            token_url = self.data_url("latest", data_item="api/token")
            register_mock_metaserver(token_url, "API-TOKEN", self.responses)
            for version in all_versions:
                metadata_url = self.data_url(version) + "/"
                if version == md_version:
                    # Register all metadata for desired version
                    register_mock_metaserver(
                        metadata_url,
                        md.get("md", DEFAULT_METADATA),
                        self.responses,
                    )
                    userdata_url = self.data_url(
                        version, data_item="user-data"
                    )
                    register_mock_metaserver(
                        userdata_url, md.get("ud", ""), self.responses
                    )
                    identity_url = self.data_url(
                        version, data_item="dynamic/instance-identity"
                    )
                    register_mock_metaserver(
                        identity_url,
                        md.get("id", DYNAMIC_METADATA),
                        self.responses,
                    )
                else:
                    instance_id_url = metadata_url + "instance-id"
                    if version == ds.min_metadata_version:
                        # Add min_metadata_version service availability check
                        register_mock_metaserver(
                            instance_id_url,
                            DEFAULT_METADATA["instance-id"],
                            self.responses,
                        )
                    else:
                        # Register 404s for all unrequested extended versions
                        register_mock_metaserver(
                            instance_id_url, None, self.responses
                        )
        return ds

    def test_network_config_property_returns_version_2_network_data(self):
        """network_config property returns network version 2 for metadata"""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": True}}},
            md={"md": DEFAULT_METADATA},
        )
        find_fallback_path = M_PATH_NET + "find_fallback_nic"
        with mock.patch(find_fallback_path) as m_find_fallback:
            m_find_fallback.return_value = "eth9"
            ds.get_data()

        mac1 = "06:17:04:d7:26:09"  # Defined in DEFAULT_METADATA
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": "06:17:04:d7:26:09"},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": True,
                }
            },
        }
        patch_path = M_PATH_NET + "get_interfaces_by_mac"
        get_interface_mac_path = M_PATH_NET + "get_interface_mac"
        with mock.patch(patch_path) as m_get_interfaces_by_mac:
            with mock.patch(find_fallback_path) as m_find_fallback:
                with mock.patch(get_interface_mac_path) as m_get_mac:
                    m_get_interfaces_by_mac.return_value = {mac1: "eth9"}
                    m_find_fallback.return_value = "eth9"
                    m_get_mac.return_value = mac1
                    self.assertEqual(expected, ds.network_config)

    def test_network_config_property_set_dhcp4(self):
        """network_config property configures dhcp4 on nics with local-ipv4s.

        Only one device is configured based on get_interfaces_by_mac even when
        multiple MACs exist in metadata.
        """
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": True}}},
            md={"md": DEFAULT_METADATA},
        )
        find_fallback_path = M_PATH_NET + "find_fallback_nic"
        with mock.patch(find_fallback_path) as m_find_fallback:
            m_find_fallback.return_value = "eth9"
            ds.get_data()

        mac1 = "06:17:04:d7:26:08"  # IPv4 only in DEFAULT_METADATA
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": mac1.lower()},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": False,
                }
            },
        }
        patch_path = M_PATH_NET + "get_interfaces_by_mac"
        get_interface_mac_path = M_PATH_NET + "get_interface_mac"
        with mock.patch(patch_path) as m_get_interfaces_by_mac:
            with mock.patch(find_fallback_path) as m_find_fallback:
                with mock.patch(get_interface_mac_path) as m_get_mac:
                    dhcp_client = ds.distro.dhcp_client
                    dhcp_client.dhcp_discovery.return_value = {
                        "routers": "172.31.1.0"
                    }
                    m_get_interfaces_by_mac.return_value = {mac1: "eth9"}
                    m_find_fallback.return_value = "eth9"
                    m_get_mac.return_value = mac1
                    self.assertEqual(expected, ds.network_config)

    def test_network_config_property_secondary_private_ips(self):
        """network_config property configures any secondary ipv4 addresses.

        Only one device is configured based on get_interfaces_by_mac even when
        multiple MACs exist in metadata.
        """
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": True}}},
            md={"md": SECONDARY_IP_METADATA_2018_09_24},
        )
        find_fallback_path = M_PATH_NET + "find_fallback_nic"
        with mock.patch(find_fallback_path) as m_find_fallback:
            m_find_fallback.return_value = "eth9"
            ds.get_data()

        mac1 = "0a:07:84:3d:6e:38"  # 1 secondary IPv4 and 2 secondary IPv6
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": mac1},
                    "set-name": "eth9",
                    "addresses": [
                        "172.31.45.70/20",
                        "2600:1f16:292:100:f152:2222:3333:4444/128",
                        "2600:1f16:292:100:f153:12a3:c37c:11f9/128",
                    ],
                    "dhcp4": True,
                    "dhcp6": True,
                }
            },
        }
        patch_path = M_PATH_NET + "get_interfaces_by_mac"
        get_interface_mac_path = M_PATH_NET + "get_interface_mac"
        with mock.patch(patch_path) as m_get_interfaces_by_mac:
            with mock.patch(find_fallback_path) as m_find_fallback:
                with mock.patch(get_interface_mac_path) as m_get_mac:
                    m_get_interfaces_by_mac.return_value = {mac1: "eth9"}
                    m_find_fallback.return_value = "eth9"
                    m_get_mac.return_value = mac1
                    self.assertEqual(expected, ds.network_config)

    def test_network_config_property_is_cached_in_datasource(self):
        """network_config property is cached in DataSourceEc2."""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": True}}},
            md={"md": DEFAULT_METADATA},
        )
        ds._network_config = {"cached": "data"}
        self.assertEqual({"cached": "data"}, ds.network_config)

    @mock.patch("cloudinit.net.dhcp.maybe_perform_dhcp_discovery")
    def test_network_config_cached_property_refreshed_on_upgrade(self, m_dhcp):
        """Refresh the network_config Ec2 cache if network key is absent.

        This catches an upgrade issue where obj.pkl contained stale metadata
        which lacked newly required network key.
        """
        old_metadata = copy.deepcopy(DEFAULT_METADATA)
        old_metadata.pop("network")
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": True}}},
            md={"md": old_metadata},
        )
        self.assertTrue(ds.get_data())

        # Workaround https://github.com/getsentry/responses/issues/212
        if hasattr(self.responses, "_urls"):
            # Can be removed when Bionic is EOL
            for index, url in enumerate(self.responses._urls):
                if url["url"].startswith(
                    "http://169.254.169.254/2009-04-04/meta-data/"
                ):
                    del self.responses._urls[index]
        elif hasattr(self.responses, "_matches"):
            # Can be removed when Focal is EOL
            for index, response in enumerate(self.responses._matches):
                if response.url.startswith(
                    "http://169.254.169.254/2009-04-04/meta-data/"
                ):
                    del self.responses._matches[index]

        # Provide new revision of metadata that contains network data
        register_mock_metaserver(
            "http://169.254.169.254/2009-04-04/meta-data/",
            DEFAULT_METADATA,
            self.responses,
        )
        mac1 = "06:17:04:d7:26:09"  # Defined in DEFAULT_METADATA
        get_interface_mac_path = M_PATH_NET + "get_interfaces_by_mac"
        ds.distro.fallback_nic = "eth9"
        with mock.patch(get_interface_mac_path) as m_get_interfaces_by_mac:
            m_get_interfaces_by_mac.return_value = {mac1: "eth9"}
            nc = ds.network_config  # Will re-crawl network metadata
            self.assertIsNotNone(nc)
        self.assertIn(
            "Refreshing stale metadata from prior to upgrade",
            self.logs.getvalue(),
        )
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": True,
                }
            },
        }
        self.assertEqual(expected, ds.network_config)

    def test_ec2_get_instance_id_refreshes_identity_on_upgrade(self):
        """get_instance-id gets DataSourceEc2Local.identity if not present.

        This handles an upgrade case where the old pickled datasource didn't
        set up self.identity, but 'systemctl cloud-init init' runs
        get_instance_id which traces on missing self.identity. lp:1748354.
        """
        self.datasource = ec2.DataSourceEc2Local
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
        )
        # Mock 404s on all versions except latest
        all_versions = [
            ds.min_metadata_version
        ] + ds.extended_metadata_versions
        for ver in all_versions[:-1]:
            register_mock_metaserver(
                "http://[fd00:ec2::254]/{0}/meta-data/instance-id".format(ver),
                None,
                self.responses,
            )

        ds.metadata_address = "http://[fd00:ec2::254]"
        register_mock_metaserver(
            "{0}/{1}/meta-data/".format(ds.metadata_address, all_versions[-1]),
            DEFAULT_METADATA,
            self.responses,
        )
        # Register dynamic/instance-identity document which we now read.
        register_mock_metaserver(
            "{0}/{1}/dynamic/".format(ds.metadata_address, all_versions[-1]),
            DYNAMIC_METADATA,
            self.responses,
        )
        ds._cloud_name = ec2.CloudNames.AWS
        # Setup cached metadata on the Datasource
        ds.metadata = DEFAULT_METADATA
        self.assertEqual("my-identity-id", ds.get_instance_id())

    def test_classic_instance_true(self):
        """If no vpc-id in metadata, is_classic_instance must return true."""
        md_copy = copy.deepcopy(DEFAULT_METADATA)
        ifaces_md = md_copy.get("network", {}).get("interfaces", {})
        for _mac, mac_data in ifaces_md.get("macs", {}).items():
            if "vpc-id" in mac_data:
                del mac_data["vpc-id"]

        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": md_copy},
        )
        self.assertTrue(ds.get_data())
        self.assertTrue(ds.is_classic_instance())

    def test_classic_instance_false(self):
        """If vpc-id in metadata, is_classic_instance must return false."""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
        )
        self.assertTrue(ds.get_data())
        self.assertFalse(ds.is_classic_instance())

    def test_aws_inaccessible_imds_service_fails_with_retries(self):
        """Inaccessibility of http://169.254.169.254 are retried."""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md=None,
        )

        conn_error = requests.exceptions.ConnectionError(
            "[Errno 113] no route to host"
        )

        mock_success = mock.MagicMock(contents=b"fakesuccess")
        mock_success.ok.return_value = True

        with mock.patch("cloudinit.url_helper.readurl") as m_readurl:
            # yikes, this endpoint needs help
            m_readurl.side_effect = (
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                conn_error,
                mock_success,
            )
            with mock.patch("cloudinit.url_helper.time.sleep"):
                self.assertTrue(ds.wait_for_metadata_service())

        # Just one /latest/api/token request
        self.assertEqual(19, len(m_readurl.call_args_list))
        for readurl_call in m_readurl.call_args_list:
            self.assertIn("latest/api/token", readurl_call[0][0])

    def test_aws_token_403_fails_without_retries(self):
        """Verify that 403s fetching AWS tokens are not retried."""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md=None,
        )

        token_url = self.data_url("latest", data_item="api/token")
        self.responses.add(responses.PUT, token_url, status=403)
        self.assertFalse(ds.get_data())
        # Just one /latest/api/token request
        logs = self.logs.getvalue()
        expected_logs = [
            "WARNING: Ec2 IMDS endpoint returned a 403 error. HTTP endpoint is"
            " disabled. Aborting.",
            "WARNING: IMDS's HTTP endpoint is probably disabled",
        ]
        for log in expected_logs:
            self.assertIn(log, logs)

    def test_aws_token_redacted(self):
        """Verify that aws tokens are redacted when logged."""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
        )
        self.assertTrue(ds.get_data())
        all_logs = self.logs.getvalue().splitlines()
        REDACT_TTL = "'X-aws-ec2-metadata-token-ttl-seconds': 'REDACTED'"
        REDACT_TOK = "'X-aws-ec2-metadata-token': 'REDACTED'"
        logs_with_redacted_ttl = [log for log in all_logs if REDACT_TTL in log]
        logs_with_redacted = [log for log in all_logs if REDACT_TOK in log]
        logs_with_token = [log for log in all_logs if "API-TOKEN" in log]
        self.assertEqual(1, len(logs_with_redacted_ttl))
        self.assertEqual(83, len(logs_with_redacted))
        self.assertEqual(0, len(logs_with_token))

    @mock.patch("cloudinit.net.dhcp.maybe_perform_dhcp_discovery")
    def test_valid_platform_with_strict_true(self, m_dhcp):
        """Valid platform data should return true with strict_id true."""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": True}}},
            md={"md": DEFAULT_METADATA},
        )
        ret = ds.get_data()
        self.assertTrue(ret)
        self.assertEqual(0, m_dhcp.call_count)
        self.assertEqual("aws", ds.cloud_name)
        self.assertEqual("ec2", ds.platform_type)
        self.assertEqual("metadata (%s)" % ds.metadata_address, ds.subplatform)

    def test_valid_platform_with_strict_false(self):
        """Valid platform data should return true with strict_id false."""
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
        )
        ret = ds.get_data()
        self.assertTrue(ret)

    def test_unknown_platform_with_strict_true(self):
        """Unknown platform data with strict_id true should return False."""
        uuid = "ab439480-72bf-11d3-91fc-b8aded755F9a"
        ds = self._setup_ds(
            platform_data={"uuid": uuid, "uuid_source": "dmi", "serial": ""},
            sys_cfg={"datasource": {"Ec2": {"strict_id": True}}},
            md={"md": DEFAULT_METADATA},
        )
        ret = ds.get_data()
        self.assertFalse(ret)

    def test_unknown_platform_with_strict_false(self):
        """Unknown platform data with strict_id false should return True."""
        uuid = "ab439480-72bf-11d3-91fc-b8aded755F9a"
        ds = self._setup_ds(
            platform_data={"uuid": uuid, "uuid_source": "dmi", "serial": ""},
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
        )
        ret = ds.get_data()
        self.assertTrue(ret)

    def test_ec2_local_returns_false_on_non_aws(self):
        """DataSourceEc2Local returns False when platform is not AWS."""
        self.datasource = ec2.DataSourceEc2Local
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
        )
        platform_attrs = [
            attr
            for attr in ec2.CloudNames.__dict__.keys()
            if not attr.startswith("__")
        ]
        for attr_name in platform_attrs:
            platform_name = getattr(ec2.CloudNames, attr_name)
            if platform_name not in ["aws", "outscale"]:
                ds._cloud_name = platform_name
                ret = ds.get_data()
                self.assertEqual("ec2", ds.platform_type)
                self.assertFalse(ret)
                message = (
                    "Local Ec2 mode only supported on ('aws', 'outscale'),"
                    " not {0}".format(platform_name)
                )
                self.assertIn(message, self.logs.getvalue())

    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_FreeBSD")
    def test_ec2_local_returns_false_on_bsd(self, m_is_freebsd):
        """DataSourceEc2Local returns False on BSD.

        FreeBSD dhclient doesn't support dhclient -sf to run in a sandbox.
        """
        m_is_freebsd.return_value = True
        self.datasource = ec2.DataSourceEc2Local
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
        )
        ret = ds.get_data()
        self.assertFalse(ret)
        self.assertIn(
            "FreeBSD doesn't support running dhclient with -sf",
            self.logs.getvalue(),
        )

    @mock.patch("cloudinit.net.ephemeral.EphemeralIPv6Network")
    @mock.patch("cloudinit.net.ephemeral.EphemeralIPv4Network")
    @mock.patch("cloudinit.distros.net.find_fallback_nic")
    @mock.patch("cloudinit.net.ephemeral.maybe_perform_dhcp_discovery")
    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_FreeBSD")
    def test_ec2_local_performs_dhcp_on_non_bsd(
        self, m_is_bsd, m_dhcp, m_fallback_nic, m_net4, m_net6
    ):
        """Ec2Local returns True for valid platform data on non-BSD with dhcp.

        DataSourceEc2Local will setup initial IPv4 network via dhcp discovery.
        Then the metadata services is crawled for more network config info.
        When the platform data is valid, return True.
        """

        m_fallback_nic.return_value = "eth9"
        m_is_bsd.return_value = False
        m_dhcp.return_value = {
            "interface": "eth9",
            "fixed-address": "192.168.2.9",
            "routers": "192.168.2.1",
            "subnet-mask": "255.255.255.0",
            "broadcast-address": "192.168.2.255",
        }
        self.datasource = ec2.DataSourceEc2Local
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": DEFAULT_METADATA},
            distro=MockDistro("", {}, {}),
        )

        ret = ds.get_data()
        self.assertTrue(ret)
        m_dhcp.assert_called_once_with(ds.distro, "eth9", None)
        m_net4.assert_called_once_with(
            ds.distro,
            broadcast="192.168.2.255",
            interface="eth9",
            ip="192.168.2.9",
            prefix_or_mask="255.255.255.0",
            router="192.168.2.1",
            static_routes=None,
        )
        self.assertIn("Crawl of metadata service ", self.logs.getvalue())

    def test_get_instance_tags(self):
        ds = self._setup_ds(
            platform_data=self.valid_platform_data,
            sys_cfg={"datasource": {"Ec2": {"strict_id": False}}},
            md={"md": TAGS_METADATA_2021_03_23},
        )
        self.assertTrue(ds.get_data())
        self.assertIn("tags", ds.metadata)
        self.assertIn("instance", ds.metadata["tags"])
        instance_tags = ds.metadata["tags"]["instance"]
        self.assertEqual(instance_tags["Application"], "test")
        self.assertEqual(instance_tags["Environment"], "production")


class TestGetSecondaryAddresses(test_helpers.CiTestCase):
    mac = "06:17:04:d7:26:ff"
    with_logs = True

    def test_md_with_no_secondary_addresses(self):
        """Empty list is returned when nic metadata contains no secondary ip"""
        self.assertEqual([], ec2.get_secondary_addresses(NIC2_MD, self.mac))

    def test_md_with_secondary_v4_and_v6_addresses(self):
        """All secondary addresses are returned from nic metadata"""
        self.assertEqual(
            [
                "172.31.45.70/20",
                "2600:1f16:292:100:f152:2222:3333:4444/128",
                "2600:1f16:292:100:f153:12a3:c37c:11f9/128",
            ],
            ec2.get_secondary_addresses(NIC1_MD_IPV4_IPV6_MULTI_IP, self.mac),
        )

    def test_invalid_ipv4_ipv6_cidr_metadata_logged_with_defaults(self):
        """Any invalid subnet-ipv(4|6)-cidr-block values use defaults"""
        invalid_cidr_md = copy.deepcopy(NIC1_MD_IPV4_IPV6_MULTI_IP)
        invalid_cidr_md["subnet-ipv4-cidr-block"] = "something-unexpected"
        invalid_cidr_md["subnet-ipv6-cidr-block"] = "not/sure/what/this/is"
        self.assertEqual(
            [
                "172.31.45.70/24",
                "2600:1f16:292:100:f152:2222:3333:4444/128",
                "2600:1f16:292:100:f153:12a3:c37c:11f9/128",
            ],
            ec2.get_secondary_addresses(invalid_cidr_md, self.mac),
        )
        expected_logs = [
            "WARNING: Could not parse subnet-ipv4-cidr-block"
            " something-unexpected for mac 06:17:04:d7:26:ff."
            " ipv4 network config prefix defaults to /24",
            "WARNING: Could not parse subnet-ipv6-cidr-block"
            " not/sure/what/this/is for mac 06:17:04:d7:26:ff."
            " ipv6 network config prefix defaults to /128",
        ]
        logs = self.logs.getvalue()
        for log in expected_logs:
            self.assertIn(log, logs)


class TestBuildNicOrder:
    @pytest.mark.parametrize(
        ["macs_metadata", "macs", "expected"],
        [
            pytest.param({}, [], {}, id="all_empty"),
            pytest.param(
                {}, ["0a:f7:8d:96:f2:a1"], {}, id="empty_macs_metadata"
            ),
            pytest.param(
                {
                    "0a:0d:dd:44:cd:7b": {
                        "device-number": "0",
                        "mac": "0a:0d:dd:44:cd:7b",
                    }
                },
                [],
                {},
                id="empty_macs",
            ),
            pytest.param(
                {
                    "0a:0d:dd:44:cd:7b": {
                        "mac": "0a:0d:dd:44:cd:7b",
                    },
                    "0a:f7:8d:96:f2:a1": {
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                },
                ["0a:f7:8d:96:f2:a1", "0a:0d:dd:44:cd:7b"],
                {"0a:f7:8d:96:f2:a1": 0, "0a:0d:dd:44:cd:7b": 1},
                id="no-device-number-info",
            ),
            pytest.param(
                {
                    "0a:0d:dd:44:cd:7b": {
                        "mac": "0a:0d:dd:44:cd:7b",
                    },
                    "0a:f7:8d:96:f2:a1": {
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                },
                ["0a:f7:8d:96:f2:a1"],
                {"0a:f7:8d:96:f2:a1": 0},
                id="no-device-number-info-subset",
            ),
            pytest.param(
                {
                    "0a:0d:dd:44:cd:7b": {
                        "device-number": "0",
                        "mac": "0a:0d:dd:44:cd:7b",
                    },
                    "0a:f7:8d:96:f2:a1": {
                        "device-number": "1",
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                },
                ["0a:f7:8d:96:f2:a1", "0a:0d:dd:44:cd:7b"],
                {"0a:0d:dd:44:cd:7b": 0, "0a:f7:8d:96:f2:a1": 1},
                id="device-numbers",
            ),
            pytest.param(
                {
                    "0a:0d:dd:44:cd:7b": {
                        "network-card": "0",
                        "device-number": "0",
                        "mac": "0a:0d:dd:44:cd:7b",
                    },
                    "0a:f7:8d:96:f2:a1": {
                        "network-card": "1",
                        "device-number": "1",
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                    "0a:f7:8d:96:f2:a2": {
                        "network-card": "2",
                        "device-number": "1",
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                },
                [
                    "0a:f7:8d:96:f2:a1",
                    "0a:0d:dd:44:cd:7b",
                    "0a:f7:8d:96:f2:a2",
                ],
                {
                    "0a:0d:dd:44:cd:7b": 0,
                    "0a:f7:8d:96:f2:a1": 1,
                    "0a:f7:8d:96:f2:a2": 2,
                },
                id="network-cardes",
            ),
            pytest.param(
                {
                    "0a:0d:dd:44:cd:7b": {
                        "network-card": "0",
                        "device-number": "0",
                        "mac": "0a:0d:dd:44:cd:7b",
                    },
                    "0a:f7:8d:96:f2:a1": {
                        "network-card": "1",
                        "device-number": "1",
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                    "0a:f7:8d:96:f2:a2": {
                        "device-number": "1",
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                },
                [
                    "0a:f7:8d:96:f2:a1",
                    "0a:0d:dd:44:cd:7b",
                    "0a:f7:8d:96:f2:a2",
                ],
                {
                    "0a:0d:dd:44:cd:7b": 0,
                    "0a:f7:8d:96:f2:a1": 1,
                    "0a:f7:8d:96:f2:a2": 2,
                },
                id="network-card-partially-missing",
            ),
            pytest.param(
                {
                    "0a:0d:dd:44:cd:7b": {
                        "mac": "0a:0d:dd:44:cd:7b",
                    },
                    "0a:f7:8d:96:f2:a1": {
                        "mac": "0a:f7:8d:96:f2:a1",
                    },
                },
                ["0a:f7:8d:96:f2:a9"],
                {},
                id="macs-not-in-md",
            ),
        ],
    )
    def test_build_nic_order(self, macs_metadata, macs, expected):
        assert expected == ec2._build_nic_order(macs_metadata, macs)


class TestConvertEc2MetadataNetworkConfig(test_helpers.CiTestCase):
    def setUp(self):
        super(TestConvertEc2MetadataNetworkConfig, self).setUp()
        self.mac1 = "06:17:04:d7:26:09"
        interface_dict = copy.deepcopy(
            DEFAULT_METADATA["network"]["interfaces"]["macs"][self.mac1]
        )
        # These tests are written assuming the base interface doesn't have IPv6
        interface_dict.pop("ipv6s")
        self.network_metadata = {
            "interfaces": {"macs": {self.mac1: interface_dict}}
        }

    def test_convert_ec2_metadata_network_config_skips_absent_macs(self):
        """Any mac absent from metadata is skipped by network config."""
        macs_to_nics = {self.mac1: "eth9", "DE:AD:BE:EF:FF:FF": "vitualnic2"}

        # DE:AD:BE:EF:FF:FF represented by OS but not in metadata
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": False,
                }
            },
        }
        distro = mock.Mock()
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                self.network_metadata, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_network_config_handles_only_dhcp6(self):
        """Config dhcp6 when ipv6s is in metadata for a mac."""
        macs_to_nics = {self.mac1: "eth9"}
        network_metadata_ipv6 = copy.deepcopy(self.network_metadata)
        nic1_metadata = network_metadata_ipv6["interfaces"]["macs"][self.mac1]
        nic1_metadata["ipv6s"] = "2620:0:1009:fd00:e442:c88d:c04d:dc85/64"
        nic1_metadata.pop("public-ipv4s")
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": True,
                }
            },
        }
        distro = mock.Mock()
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                network_metadata_ipv6, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_network_config_local_only_dhcp4(self):
        """Config dhcp4 when there are no public addresses in public-ipv4s."""
        macs_to_nics = {self.mac1: "eth9"}
        network_metadata_ipv6 = copy.deepcopy(self.network_metadata)
        nic1_metadata = network_metadata_ipv6["interfaces"]["macs"][self.mac1]
        nic1_metadata["local-ipv4s"] = "172.3.3.15"
        nic1_metadata.pop("public-ipv4s")
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": False,
                }
            },
        }
        distro = mock.Mock()
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                network_metadata_ipv6, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_network_config_handles_absent_dhcp4(self):
        """Config dhcp4 on fallback_nic when there are no ipv4 addresses."""
        macs_to_nics = {self.mac1: "eth9"}
        network_metadata_ipv6 = copy.deepcopy(self.network_metadata)
        nic1_metadata = network_metadata_ipv6["interfaces"]["macs"][self.mac1]
        nic1_metadata["public-ipv4s"] = ""

        # When no ipv4 or ipv6 content but fallback_nic set, set dhcp4 config.
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": False,
                }
            },
        }
        distro = mock.Mock()
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                network_metadata_ipv6, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_network_config_handles_local_v4_and_v6(self):
        """When ipv6s and local-ipv4s are non-empty, enable dhcp6 and dhcp4."""
        macs_to_nics = {self.mac1: "eth9"}
        network_metadata_both = copy.deepcopy(self.network_metadata)
        nic1_metadata = network_metadata_both["interfaces"]["macs"][self.mac1]
        nic1_metadata["ipv6s"] = "2620:0:1009:fd00:e442:c88d:c04d:dc85/64"
        nic1_metadata.pop("public-ipv4s")
        nic1_metadata["local-ipv4s"] = "10.0.0.42"  # Local ipv4 only on vpc
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": True,
                }
            },
        }
        distro = mock.Mock()
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                network_metadata_both, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_network_config_multi_nics_ipv4(self):
        """DHCP route-metric increases on secondary NICs for IPv4 and IPv6.
        Source-routing configured for secondary NICs (routing-policy and extra
        routing table)."""
        mac2 = "06:17:04:d7:26:08"
        macs_to_nics = {self.mac1: "eth9", mac2: "eth10"}
        network_metadata_both = copy.deepcopy(self.network_metadata)
        # Add 2nd nic info
        network_metadata_both["interfaces"]["macs"][mac2] = NIC2_MD
        nic1_metadata = network_metadata_both["interfaces"]["macs"][self.mac1]
        nic1_metadata["ipv6s"] = "2620:0:1009:fd00:e442:c88d:c04d:dc85/64"
        nic1_metadata.pop("public-ipv4s")  # No public-ipv4 IPs in cfg
        nic1_metadata["local-ipv4s"] = "10.0.0.42"  # Local ipv4 only on vpc
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp4-overrides": {"route-metric": 100},
                    "dhcp6": True,
                    "dhcp6-overrides": {"route-metric": 100},
                },
                "eth10": {
                    "match": {"macaddress": mac2},
                    "set-name": "eth10",
                    "dhcp4": True,
                    "dhcp4-overrides": {
                        "route-metric": 200,
                        "use-routes": True,
                    },
                    "dhcp6": False,
                    "routes": [
                        # via DHCP gateway
                        {"to": "0.0.0.0/0", "via": "172.31.1.0", "table": 101},
                        # to NIC2_MD["subnet-ipv4-cidr-block"]
                        {"to": "172.31.32.0/20", "table": 101},
                    ],
                    "routing-policy": [
                        # NIC2_MD["local-ipv4s"]
                        {"from": "172.31.47.221", "table": 101}
                    ],
                },
            },
        }
        distro = mock.Mock()
        distro.network_activator = activators.NetplanActivator
        distro.dhcp_client.dhcp_discovery.return_value = {
            "routers": "172.31.1.0"
        }
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                network_metadata_both, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_network_config_multi_nics_ipv4_ipv6_multi_ip(
        self,
    ):
        """DHCP route-metric increases on secondary NICs for IPv4 and IPv6.
        Source-routing configured for secondary NICs (routing-policy and extra
        routing table)."""
        mac2 = "06:17:04:d7:26:08"
        macs_to_nics = {self.mac1: "eth9", mac2: "eth10"}
        network_metadata_both = copy.deepcopy(self.network_metadata)
        # Add 2nd nic info
        network_metadata_both["interfaces"]["macs"][
            mac2
        ] = NIC2_MD_IPV4_IPV6_MULTI_IP
        nic1_metadata = network_metadata_both["interfaces"]["macs"][self.mac1]
        nic1_metadata["ipv6s"] = "2620:0:1009:fd00:e442:c88d:c04d:dc85/64"
        nic1_metadata.pop("public-ipv4s")  # No public-ipv4 IPs in cfg
        nic1_metadata["local-ipv4s"] = "10.0.0.42"  # Local ipv4 only on vpc
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "dhcp4": True,
                    "dhcp4-overrides": {"route-metric": 100},
                    "dhcp6": True,
                    "match": {"macaddress": "06:17:04:d7:26:09"},
                    "set-name": "eth9",
                    "dhcp6-overrides": {"route-metric": 100},
                },
                "eth10": {
                    "dhcp4": True,
                    "dhcp4-overrides": {
                        "route-metric": 200,
                        "use-routes": True,
                    },
                    "dhcp6": True,
                    "match": {"macaddress": "06:17:04:d7:26:08"},
                    "set-name": "eth10",
                    "routes": [
                        # via DHCP gateway
                        {"to": "0.0.0.0/0", "via": "172.31.1.0", "table": 101},
                        # to NIC2_MD["subnet-ipv4-cidr-block"]
                        {"to": "172.31.32.0/20", "table": 101},
                        # to NIC2_MD["subnet-ipv6-cidr-blocks"]
                        {"to": "2600:1f16:292:100::/64", "table": 101},
                    ],
                    "routing-policy": [
                        # NIC2_MD["local-ipv4s"]
                        {"from": "172.31.47.221", "table": 101},
                        {
                            "from": "2600:1f16:292:100:c187:593c:4349:136",
                            "table": 101,
                        },
                        {
                            "from": "2600:1f16:292:100:f153:12a3:c37c:11f9",
                            "table": 101,
                        },
                    ],
                    "dhcp6-overrides": {
                        "route-metric": 200,
                        "use-routes": True,
                    },
                    "addresses": ["2600:1f16:292:100:f153:12a3:c37c:11f9/128"],
                },
            },
        }
        distro = mock.Mock()
        distro.network_activator = activators.NetplanActivator
        distro.dhcp_client.dhcp_discovery.return_value = {
            "routers": "172.31.1.0"
        }
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                network_metadata_both, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_network_config_handles_dhcp4_and_dhcp6(self):
        """Config both dhcp4 and dhcp6 when both vpc-ipv6 and ipv4 exists."""
        macs_to_nics = {self.mac1: "eth9"}
        network_metadata_both = copy.deepcopy(self.network_metadata)
        nic1_metadata = network_metadata_both["interfaces"]["macs"][self.mac1]
        nic1_metadata["ipv6s"] = "2620:0:1009:fd00:e442:c88d:c04d:dc85/64"
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": True,
                }
            },
        }
        distro = mock.Mock()
        self.assertEqual(
            expected,
            ec2.convert_ec2_metadata_network_config(
                network_metadata_both, distro, macs_to_nics
            ),
        )

    def test_convert_ec2_metadata_gets_macs_from_get_interfaces_by_mac(self):
        """Convert Ec2 Metadata calls get_interfaces_by_mac by default."""
        expected = {
            "version": 2,
            "ethernets": {
                "eth9": {
                    "match": {"macaddress": self.mac1},
                    "set-name": "eth9",
                    "dhcp4": True,
                    "dhcp6": False,
                }
            },
        }
        patch_path = M_PATH_NET + "get_interfaces_by_mac"
        distro = mock.Mock()
        with mock.patch(patch_path) as m_get_interfaces_by_mac:
            m_get_interfaces_by_mac.return_value = {self.mac1: "eth9"}
            self.assertEqual(
                expected,
                ec2.convert_ec2_metadata_network_config(
                    self.network_metadata, distro
                ),
            )


class TesIdentifyPlatform(test_helpers.CiTestCase):
    def collmock(self, **kwargs):
        """return non-special _collect_platform_data updated with changes."""
        unspecial = {
            "asset_tag": "3857-0037-2746-7462-1818-3997-77",
            "serial": "H23-C4J3JV-R6",
            "uuid": "81c7e555-6471-4833-9551-1ab366c4cfd2",
            "uuid_source": "dmi",
            "vendor": "tothecloud",
            "product_name": "cloudproduct",
        }
        unspecial.update(**kwargs)
        return unspecial

    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_identify_aliyun(self, m_collect):
        """aliyun should be identified if product name equals to
        Alibaba Cloud ECS
        """
        m_collect.return_value = self.collmock(
            product_name="Alibaba Cloud ECS"
        )
        self.assertEqual(ec2.CloudNames.ALIYUN, ec2.identify_platform())

    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_identify_zstack(self, m_collect):
        """zstack should be identified if chassis-asset-tag
        ends in .zstack.io
        """
        m_collect.return_value = self.collmock(asset_tag="123456.zstack.io")
        self.assertEqual(ec2.CloudNames.ZSTACK, ec2.identify_platform())

    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_identify_zstack_full_domain_only(self, m_collect):
        """zstack asset-tag matching should match only on
        full domain boundary.
        """
        m_collect.return_value = self.collmock(asset_tag="123456.buzzstack.io")
        self.assertEqual(ec2.CloudNames.UNKNOWN, ec2.identify_platform())

    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_identify_e24cloud(self, m_collect):
        """e24cloud identified if vendor is e24cloud"""
        m_collect.return_value = self.collmock(vendor="e24cloud")
        self.assertEqual(ec2.CloudNames.E24CLOUD, ec2.identify_platform())

    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_identify_e24cloud_negative(self, m_collect):
        """e24cloud identified if vendor is e24cloud"""
        m_collect.return_value = self.collmock(vendor="e24cloudyday")
        self.assertEqual(ec2.CloudNames.UNKNOWN, ec2.identify_platform())

    # Outscale
    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_identify_outscale(self, m_collect):
        """Should return true if the dmi product data has expected value."""
        m_collect.return_value = self.collmock(
            vendor="3DS Outscale".lower(),
            product_name="3DS Outscale VM".lower(),
        )
        self.assertEqual(ec2.CloudNames.OUTSCALE, ec2.identify_platform())

    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_false_on_wrong_sys_vendor(self, m_collect):
        """Should return false on empty value returned."""
        m_collect.return_value = self.collmock(
            vendor="Not 3DS Outscale".lower(),
            product_name="3DS Outscale VM".lower(),
        )
        self.assertEqual(ec2.CloudNames.UNKNOWN, ec2.identify_platform())

    @mock.patch("cloudinit.sources.DataSourceEc2._collect_platform_data")
    def test_false_on_wrong_product_name(self, m_collect):
        """Should return false on an unrelated string."""
        m_collect.return_value = self.collmock(
            vendor="3DS Outscale".lower(),
            product_name="Not 3DS Outscale VM".lower(),
        )
        self.assertEqual(ec2.CloudNames.UNKNOWN, ec2.identify_platform())



import pytest

USER_DATA = """\
#cloud-config
keyboard:
    layout: de
    model: pc105
    variant: nodeadkeys
    options: compose:rwin
"""


class TestKeyboard:
    @pytest.mark.user_data(USER_DATA)
    def test_keyboard(self, client):
        lc = client.execute("localectl")
        assert "X11 Layout: de" in lc



# This file is part of cloud-init. See LICENSE file for license information.

import os
import shutil
import tempfile
from unittest import mock

import pytest

from cloudinit import distros, util
from cloudinit.distros.ubuntu import Distro
from cloudinit.net.dhcp import Dhcpcd, IscDhclient, Udhcpc
from tests.unittests import helpers

M_PATH = "cloudinit.distros."

unknown_arch_info = {
    "arches": ["default"],
    "failsafe": {
        "primary": "http://fs-primary-default",
        "security": "http://fs-security-default",
    },
}

package_mirrors = [
    {
        "arches": ["i386", "amd64"],
        "failsafe": {
            "primary": "http://fs-primary-intel",
            "security": "http://fs-security-intel",
        },
        "search": {
            "primary": [
                "http://%(ec2_region)s.ec2/",
                "http://%(availability_zone)s.clouds/",
            ],
            "security": [
                "http://security-mirror1-intel",
                "http://security-mirror2-intel",
            ],
        },
    },
    {
        "arches": ["armhf", "armel"],
        "failsafe": {
            "primary": "http://fs-primary-arm",
            "security": "http://fs-security-arm",
        },
    },
    unknown_arch_info,
]

gpmi = distros._get_package_mirror_info
gapmi = distros._get_arch_package_mirror_info


class TestGenericDistro(helpers.FilesystemMockingTestCase):
    def setUp(self):
        super(TestGenericDistro, self).setUp()
        # Make a temp directoy for tests to use.
        self.tmp = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmp)

    def _write_load_doas(self, user, rules):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        if not os.path.exists(os.path.join(self.tmp, "etc")):
            os.makedirs(os.path.join(self.tmp, "etc"))
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        d.write_doas_rules(user, rules)
        contents = util.load_text_file(d.doas_fn)
        return contents, cls, d

    def _write_load_sudoers(self, _user, rules):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        os.makedirs(os.path.join(self.tmp, "etc"))
        os.makedirs(os.path.join(self.tmp, "etc", "sudoers.d"))
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        d.write_sudo_rules("harlowja", rules)
        contents = util.load_text_file(d.ci_sudoers_fn)
        return contents, cls, d

    def _count_in(self, lines_look_for, text_content):
        found_amount = 0
        for e in lines_look_for:
            for line in text_content.splitlines():
                line = line.strip()
                if line == e:
                    found_amount += 1
        return found_amount

    def test_doas_ensure_rules(self):
        rules = ["permit nopass harlowja"]
        contents = self._write_load_doas("harlowja", rules)[0]
        expected = ["permit nopass harlowja"]
        self.assertEqual(len(expected), self._count_in(expected, contents))

    def test_doas_ensure_rules_list(self):
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        contents = self._write_load_doas("harlowja", rules)[0]
        expected = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        self.assertEqual(len(expected), self._count_in(expected, contents))

    def test_doas_ensure_handle_duplicates(self):
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        d = self._write_load_doas("harlowja", rules)[2]
        # write to doas.conf again - should not create duplicate rules
        d.write_doas_rules("harlowja", rules)
        contents = util.load_text_file(d.doas_fn)
        expected = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        self.assertEqual(len(expected), self._count_in(expected, contents))

    def test_doas_ensure_new(self):
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        contents = self._write_load_doas("harlowja", rules)[0]
        self.assertIn("# Created by cloud-init v.", contents)
        self.assertIn("harlowja", contents)
        self.assertEqual(4, contents.count("harlowja"))

    def test_doas_ensure_append(self):
        self.patchUtils(self.tmp)
        util.write_file("/etc/doas.conf", "# root user\npermit nopass root\n")
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        contents = self._write_load_doas("harlowja", rules)[0]
        self.assertIn("root", contents)
        self.assertEqual(2, contents.count("root"))
        self.assertIn("harlowja", contents)
        self.assertEqual(4, contents.count("harlowja"))

    def test_sudoers_ensure_rules(self):
        rules = "ALL=(ALL:ALL) ALL"
        contents = self._write_load_sudoers("harlowja", rules)[0]
        expected = ["harlowja ALL=(ALL:ALL) ALL"]
        self.assertEqual(len(expected), self._count_in(expected, contents))
        not_expected = [
            "harlowja A",
            "harlowja L",
            "harlowja L",
        ]
        self.assertEqual(0, self._count_in(not_expected, contents))

    def test_sudoers_ensure_rules_list(self):
        rules = [
            "ALL=(ALL:ALL) ALL",
            "B-ALL=(ALL:ALL) ALL",
            "C-ALL=(ALL:ALL) ALL",
        ]
        contents = self._write_load_sudoers("harlowja", rules)[0]
        expected = [
            "harlowja ALL=(ALL:ALL) ALL",
            "harlowja B-ALL=(ALL:ALL) ALL",
            "harlowja C-ALL=(ALL:ALL) ALL",
        ]
        self.assertEqual(len(expected), self._count_in(expected, contents))
        not_expected = [
            "harlowja A",
            "harlowja L",
            "harlowja L",
        ]
        self.assertEqual(0, self._count_in(not_expected, contents))

    def test_sudoers_ensure_handle_duplicates(self):
        rules = [
            "ALL=(ALL:ALL) ALL",
            "B-ALL=(ALL:ALL) ALL",
            "C-ALL=(ALL:ALL) ALL",
        ]
        d = self._write_load_sudoers("harlowja", rules)[2]
        # write to sudoers again - should not create duplicate rules
        d.write_sudo_rules("harlowja", rules)
        contents = util.load_text_file(d.ci_sudoers_fn)
        expected = [
            "harlowja ALL=(ALL:ALL) ALL",
            "harlowja B-ALL=(ALL:ALL) ALL",
            "harlowja C-ALL=(ALL:ALL) ALL",
        ]
        self.assertEqual(len(expected), self._count_in(expected, contents))
        not_expected = [
            "harlowja A",
            "harlowja L",
            "harlowja L",
        ]
        self.assertEqual(0, self._count_in(not_expected, contents))

    def test_sudoers_ensure_new(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        d.ensure_sudo_dir("/b")
        contents = util.load_text_file("/etc/sudoers")
        self.assertIn("includedir /b", contents)
        self.assertTrue(os.path.isdir("/b"))

    def test_sudoers_ensure_append(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        util.write_file("/etc/sudoers", "josh, josh\n")
        d.ensure_sudo_dir("/b")
        contents = util.load_text_file("/etc/sudoers")
        self.assertIn("includedir /b", contents)
        self.assertTrue(os.path.isdir("/b"))
        self.assertIn("josh", contents)
        self.assertEqual(2, contents.count("josh"))

    def test_sudoers_ensure_only_one_includedir(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        for char in ["#", "@"]:
            util.write_file("/etc/sudoers", "{}includedir /b".format(char))
            d.ensure_sudo_dir("/b")
            contents = util.load_text_file("/etc/sudoers")
            self.assertIn("includedir /b", contents)
            self.assertTrue(os.path.isdir("/b"))
            self.assertEqual(1, contents.count("includedir /b"))

    def test_arch_package_mirror_info_unknown(self):
        """for an unknown arch, we should get back that with arch 'default'."""
        arch_mirrors = gapmi(package_mirrors, arch="unknown")
        self.assertEqual(unknown_arch_info, arch_mirrors)

    def test_arch_package_mirror_info_known(self):
        arch_mirrors = gapmi(package_mirrors, arch="amd64")
        self.assertEqual(package_mirrors[0], arch_mirrors)

    def test_systemd_in_use(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        os.makedirs("/run/systemd/system")
        self.assertTrue(d.uses_systemd())

    def test_systemd_not_in_use(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        self.assertFalse(d.uses_systemd())

    def test_systemd_symlink(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        self.patchOS(self.tmp)
        self.patchUtils(self.tmp)
        os.makedirs("/run/systemd")
        os.symlink("/", "/run/systemd/system")
        self.assertFalse(d.uses_systemd())

    @mock.patch("cloudinit.distros.debian.read_system_locale")
    def test_get_locale_ubuntu(self, m_locale):
        """Test ubuntu distro returns locale set to C.UTF-8"""
        m_locale.return_value = "C.UTF-8"
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        locale = d.get_locale()
        self.assertEqual("C.UTF-8", locale)

    @mock.patch("cloudinit.distros.rhel.Distro._read_system_locale")
    def test_get_locale_rhel(self, m_locale):
        """Test rhel distro returns locale set to C.UTF-8"""
        m_locale.return_value = "C.UTF-8"
        cls = distros.fetch("rhel")
        d = cls("rhel", {}, None)
        locale = d.get_locale()
        self.assertEqual("C.UTF-8", locale)

    def test_expire_passwd_uses_chpasswd(self):
        """Test ubuntu.expire_passwd uses the passwd command."""
        for d_name in ("ubuntu", "rhel"):
            cls = distros.fetch(d_name)
            d = cls(d_name, {}, None)
            with mock.patch("cloudinit.subp.subp") as m_subp:
                d.expire_passwd("myuser")
            m_subp.assert_called_once_with(["passwd", "--expire", "myuser"])

    def test_expire_passwd_freebsd_uses_pw_command(self):
        """Test FreeBSD.expire_passwd uses the pw command."""
        cls = distros.fetch("freebsd")
        # patch ifconfig -a
        with mock.patch(
            "cloudinit.distros.networking.subp.subp", return_value=("", None)
        ):
            d = cls("freebsd", {}, None)
        with mock.patch("cloudinit.subp.subp") as m_subp:
            d.expire_passwd("myuser")
        m_subp.assert_called_once_with(
            ["pw", "usermod", "myuser", "-p", "01-Jan-1970"]
        )


class TestGetPackageMirrors:
    def return_first(self, mlist):
        if not mlist:
            return None
        return mlist[0]

    def return_second(self, mlist):
        if not mlist:
            return None

        return mlist[1] if len(mlist) > 1 else None

    def return_none(self, _mlist):
        return None

    def return_last(self, mlist):
        if not mlist:
            return None
        return mlist[-1]

    @pytest.mark.parametrize(
        "allow_ec2_mirror, platform_type, mirrors",
        [
            (
                True,
                "ec2",
                [
                    {
                        "primary": "http://us-east-1.ec2/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
            (
                True,
                "other",
                [
                    {
                        "primary": "http://us-east-1.ec2/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
            (
                False,
                "ec2",
                [
                    {
                        "primary": "http://us-east-1.ec2/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
            (
                False,
                "other",
                [
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://fs-primary-intel",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
        ],
    )
    def test_get_package_mirror_info_az_ec2(
        self, allow_ec2_mirror, platform_type, mirrors
    ):
        flag_path = (
            "cloudinit.distros.ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES"
        )
        with mock.patch(flag_path, allow_ec2_mirror):
            arch_mirrors = gapmi(package_mirrors, arch="amd64")
            data_source_mock = mock.Mock(
                availability_zone="us-east-1a", platform_type=platform_type
            )

            results = gpmi(
                arch_mirrors,
                data_source=data_source_mock,
                mirror_filter=self.return_first,
            )
            assert results == mirrors[0]

            results = gpmi(
                arch_mirrors,
                data_source=data_source_mock,
                mirror_filter=self.return_second,
            )
            assert results == mirrors[1]

            results = gpmi(
                arch_mirrors,
                data_source=data_source_mock,
                mirror_filter=self.return_none,
            )
            assert results == package_mirrors[0]["failsafe"]

    def test_get_package_mirror_info_az_non_ec2(self):
        arch_mirrors = gapmi(package_mirrors, arch="amd64")
        data_source_mock = mock.Mock(availability_zone="nova.cloudvendor")

        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_first,
        )
        assert results == {
            "primary": "http://nova.cloudvendor.clouds/",
            "security": "http://security-mirror1-intel",
        }

        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_last,
        )
        assert results == {
            "primary": "http://nova.cloudvendor.clouds/",
            "security": "http://security-mirror2-intel",
        }

    def test_get_package_mirror_info_none(self):
        arch_mirrors = gapmi(package_mirrors, arch="amd64")
        data_source_mock = mock.Mock(availability_zone=None)

        # because both search entries here replacement based on
        # availability-zone, the filter will be called with an empty list and
        # failsafe should be taken.
        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_first,
        )
        assert results == {
            "primary": "http://fs-primary-intel",
            "security": "http://security-mirror1-intel",
        }

        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_last,
        )
        assert results == {
            "primary": "http://fs-primary-intel",
            "security": "http://security-mirror2-intel",
        }


@pytest.mark.usefixtures("fake_filesystem")
class TestDistro:
    @pytest.mark.parametrize("is_noexec", [False, True])
    @mock.patch(M_PATH + "util.has_mount_opt")
    @mock.patch(M_PATH + "temp_utils.get_tmp_ancestor", return_value="/tmp")
    def test_get_tmp_exec_path(
        self, m_get_tmp_ancestor, m_has_mount_opt, is_noexec, mocker
    ):
        m_has_mount_opt.return_value = not is_noexec
        cls = distros.fetch("ubuntu")
        distro = cls("ubuntu", {}, None)
        mocker.patch.object(distro, "usr_lib_exec", "/usr_lib_exec")
        tmp_path = distro.get_tmp_exec_path()
        if is_noexec:
            assert "/tmp" == tmp_path
        else:
            assert "/usr_lib_exec/cloud-init/clouddir" == tmp_path


@pytest.mark.parametrize(
    "chosen_client, config, which_override",
    [
        pytest.param(
            IscDhclient,
            {"network": {"dhcp_client_priority": ["dhclient"]}},
            None,
            id="single_client_is_found_from_config_dhclient",
        ),
        pytest.param(
            Udhcpc,
            {"network": {"dhcp_client_priority": ["udhcpc"]}},
            None,
            id="single_client_is_found_from_config_udhcpc",
        ),
        pytest.param(
            Dhcpcd,
            {"network": {"dhcp_client_priority": ["dhcpcd"]}},
            None,
            id="single_client_is_found_from_config_dhcpcd",
        ),
        pytest.param(
            Dhcpcd,
            {"network": {"dhcp_client_priority": ["dhcpcd", "dhclient"]}},
            None,
            id="first_client_is_found_from_config_dhcpcd",
        ),
        pytest.param(
            Udhcpc,
            {
                "network": {
                    "dhcp_client_priority": ["udhcpc", "dhcpcd", "dhclient"]
                }
            },
            None,
            id="first_client_is_found_from_config_udhcpc",
        ),
        pytest.param(
            Dhcpcd,
            {"network": {"dhcp_client_priority": []}},
            None,
            id="first_client_is_found_no_config_dhcpcd",
        ),
        pytest.param(
            Dhcpcd,
            {
                "network": {
                    "dhcp_client_priority": ["udhcpc", "dhcpcd", "dhclient"]
                }
            },
            [False, False, True, True],
            id="second_client_is_found_from_config_dhcpcd",
        ),
    ],
)
class TestDHCP:
    @mock.patch("cloudinit.net.dhcp.subp.which")
    def test_dhcp_configuration(
        self, m_which, chosen_client, config, which_override
    ):
        """check that, when a user provides a configuration at
        network.dhcp_client_priority, the correct client is chosen
        """
        m_which.side_effect = which_override
        distro = Distro("", {}, {})
        distro._cfg = config
        assert isinstance(distro.dhcp_client, chosen_client)



# Copyright (C) 2019 Canonical Ltd.
# Copyright (C) 2019 VMware INC.
#
# Author: Xiaofeng Wang <xiaofengw@vmware.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit import subp
from cloudinit.sources.helpers.vmware.imc.config import Config
from cloudinit.sources.helpers.vmware.imc.config_file import ConfigFile
from cloudinit.sources.helpers.vmware.imc.guestcust_util import (
    get_tools_config,
    set_gc_status,
)
from cloudinit.subp import SubpResult
from tests.unittests.helpers import CiTestCase, mock


class TestGuestCustUtil(CiTestCase):
    def test_get_tools_config_not_installed(self):
        """
        This test is designed to verify the behavior if vmware-toolbox-cmd
        is not installed.
        """
        with mock.patch.object(subp, "which", return_value=None):
            self.assertEqual(
                get_tools_config("section", "key", "defaultVal"), "defaultVal"
            )

    def test_get_tools_config_internal_exception(self):
        """
        This test is designed to verify the behavior if internal exception
        is raised.
        """
        with mock.patch.object(subp, "which", return_value="/dummy/path"):
            with mock.patch.object(
                subp,
                "subp",
                return_value=SubpResult("key=value", b""),
                side_effect=subp.ProcessExecutionError(
                    "subp failed", exit_code=99
                ),
            ):
                # verify return value is 'defaultVal', not 'value'.
                self.assertEqual(
                    get_tools_config("section", "key", "defaultVal"),
                    "defaultVal",
                )

    def test_get_tools_config_normal(self):
        """
        This test is designed to verify the value could be parsed from
        key = value of the given [section]
        """
        with mock.patch.object(subp, "which", return_value="/dummy/path"):
            # value is not blank
            with mock.patch.object(
                subp, "subp", return_value=SubpResult("key =   value  ", b"")
            ):
                self.assertEqual(
                    get_tools_config("section", "key", "defaultVal"), "value"
                )
            # value is blank
            with mock.patch.object(
                subp, "subp", return_value=SubpResult("key = ", b"")
            ):
                self.assertEqual(
                    get_tools_config("section", "key", "defaultVal"), ""
                )
            # value contains =
            with mock.patch.object(
                subp, "subp", return_value=SubpResult("key=Bar=Wark", b"")
            ):
                self.assertEqual(
                    get_tools_config("section", "key", "defaultVal"),
                    "Bar=Wark",
                )

            # value contains specific characters
            with mock.patch.object(
                subp, "subp", return_value=SubpResult("[a] b.c_d=e-f", b"")
            ):
                self.assertEqual(
                    get_tools_config("section", "key", "defaultVal"), "e-f"
                )

    def test_set_gc_status(self):
        """
        This test is designed to verify the behavior of set_gc_status
        """
        # config is None, return None
        self.assertEqual(set_gc_status(None, "Successful"), None)

        # post gc status is NO, return None
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        self.assertEqual(set_gc_status(conf, "Successful"), None)

        # post gc status is YES, subp is called to execute command
        cf._insertKey("MISC|POST-GC-STATUS", "YES")
        conf = Config(cf)
        with mock.patch.object(
            subp, "subp", return_value=SubpResult("ok", b"")
        ) as mockobj:
            self.assertEqual(set_gc_status(conf, "Successful"), ("ok", b""))
            mockobj.assert_called_once_with(
                ["vmware-rpctool", "info-set guestinfo.gc.status Successful"],
                rcs=[0],
            )



# Copyright (C) 2009-2010 Canonical Ltd.
# Copyright (C) 2012, 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
"""SSH Import ID: Import SSH id"""

import logging
import pwd
from textwrap import dedent

from cloudinit import subp, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.distros import ug_util
from cloudinit.settings import PER_INSTANCE

# https://launchpad.net/ssh-import-id
distros = ["alpine", "cos", "debian", "ubuntu"]

SSH_IMPORT_ID_BINARY = "ssh-import-id"
MODULE_DESCRIPTION = """\
This module imports SSH keys from either a public keyserver, usually launchpad
or github using ``ssh-import-id``. Keys are referenced by the username they are
associated with on the keyserver. The keyserver can be specified by prepending
either ``lp:`` for launchpad or ``gh:`` for github to the username.
"""

meta: MetaSchema = {
    "id": "cc_ssh_import_id",
    "name": "SSH Import ID",
    "title": "Import SSH id",
    "description": MODULE_DESCRIPTION,
    "distros": distros,
    "frequency": PER_INSTANCE,
    "examples": [
        dedent(
            """\
            ssh_import_id:
             - user
             - gh:user
             - lp:user
            """
        )
    ],
    "activate_by_schema_keys": [],
}

__doc__ = get_meta_doc(meta)
LOG = logging.getLogger(__name__)


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:

    if not is_key_in_nested_dict(cfg, "ssh_import_id"):
        LOG.debug(
            "Skipping module named ssh_import_id, no 'ssh_import_id'"
            " directives found."
        )
        return
    elif not subp.which(SSH_IMPORT_ID_BINARY):
        LOG.warning(
            "ssh-import-id is not installed, but module ssh_import_id is "
            "configured. Skipping module."
        )
        return

    # import for "user: XXXXX"
    if len(args) != 0:
        user = args[0]
        ids = []
        if len(args) > 1:
            ids = args[1:]

        import_ssh_ids(ids, user)
        return

    # import for cloudinit created users
    (users, _groups) = ug_util.normalize_users_groups(cfg, cloud.distro)
    elist = []
    for (user, user_cfg) in users.items():
        import_ids = []
        if user_cfg["default"]:
            import_ids = util.get_cfg_option_list(cfg, "ssh_import_id", [])
        else:
            try:
                import_ids = user_cfg["ssh_import_id"]
            except Exception:
                LOG.debug("User %s is not configured for ssh_import_id", user)
                continue

        try:
            import_ids = util.uniq_merge(import_ids)
            import_ids = [str(i) for i in import_ids]
        except Exception:
            LOG.debug(
                "User %s is not correctly configured for ssh_import_id", user
            )
            continue

        if not len(import_ids):
            continue

        try:
            import_ssh_ids(import_ids, user)
        except Exception as exc:
            util.logexc(
                LOG, "ssh-import-id failed for: %s %s", user, import_ids
            )
            elist.append(exc)

    if len(elist):
        raise elist[0]


def import_ssh_ids(ids, user):

    if not (user and ids):
        LOG.debug("empty user(%s) or ids(%s). not importing", user, ids)
        return

    try:
        pwd.getpwnam(user)
    except KeyError as exc:
        raise exc

    # TODO: We have a use case that involes setting a proxy value earlier
    # in boot and the user wants this env used when using ssh-import-id.
    # E.g.,:
    # bootcmd:
    #   - mkdir -p /etc/systemd/system/cloud-config.service.d
    #   - mkdir -p /etc/systemd/system/cloud-final.service.d
    # write_files:
    #   - content: |
    #       http_proxy=http://192.168.1.2:3128/
    #       https_proxy=http://192.168.1.2:3128/
    #     path: /etc/cloud/env
    #   - content: |
    #       [Service]
    #       EnvironmentFile=/etc/cloud/env
    #       PassEnvironment=https_proxy http_proxy
    #     path: /etc/systemd/system/cloud-config.service.d/override.conf
    #   - content: |
    #       [Service]
    #       EnvironmentFile=/etc/cloud/env
    #       PassEnvironment=https_proxy http_proxy
    #     path: /etc/systemd/system/cloud-final.service.d/override.conf
    #
    # I'm including the `--preserve-env` here as a one-off, but we should
    # have a better way of setting env earlier in boot and using it later.
    # Perhaps a 'set_env' module?
    if subp.which("sudo"):
        cmd = [
            "sudo",
            "--preserve-env=https_proxy",
            "-Hu",
            user,
            SSH_IMPORT_ID_BINARY,
        ] + ids
    elif subp.which("doas"):
        cmd = [
            "doas",
            "-u",
            user,
            SSH_IMPORT_ID_BINARY,
        ] + ids
    else:
        LOG.error("Neither sudo nor doas available! Unable to import SSH ids.")
        return
    LOG.debug("Importing SSH ids for user %s.", user)

    try:
        subp.subp(cmd, capture=False)
    except subp.ProcessExecutionError as exc:
        util.logexc(LOG, "Failed to run command to import %s SSH ids", user)
        raise exc


def is_key_in_nested_dict(config: dict, search_key: str) -> bool:
    """Search for key nested in config.

    Note: A dict embedded in a list of lists will not be found walked - but in
    this case we don't need it.
    """
    for config_key in config.keys():
        if search_key == config_key:
            return True
        if isinstance(config[config_key], dict):
            if is_key_in_nested_dict(config[config_key], search_key):
                return True
        if isinstance(config[config_key], list):
            # this code could probably be generalized to walking the whole
            # config by iterating lists in search of dictionaries
            for item in config[config_key]:
                if isinstance(item, dict):
                    if is_key_in_nested_dict(item, search_key):
                        return True
    return False



import json
import logging
import os

import pytest
from pycloudlib.cloud import ImageType

from tests.integration_tests.clouds import IntegrationCloud
from tests.integration_tests.conftest import get_validated_source
from tests.integration_tests.instances import (
    CloudInitSource,
    IntegrationInstance,
)
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.releases import (
    BIONIC,
    CURRENT_RELEASE,
    FOCAL,
    IS_UBUNTU,
    JAMMY,
)
from tests.integration_tests.util import verify_clean_log

LOG = logging.getLogger("integration_testing.test_ubuntu_pro")

CLOUD_INIT_UA_TOKEN = os.environ.get("CLOUD_INIT_UA_TOKEN")

ATTACH_FALLBACK = """\
#cloud-config
ubuntu_pro:
  features:
    disable_auto_attach: true
  token: {token}
"""

ATTACH = """\
#cloud-config
ubuntu_pro:
  token: {token}
  enable:
  - esm-infra
"""

PRO_AUTO_ATTACH_DISABLED = """\
#cloud-config
# ubuntu_advantage config kept as duplication until the release of this
# commit in proclient (ubuntu-advantage-tools v. 32):
# https://github.com/canonical/ubuntu-pro-client/commit/7bb69e3ad
# Without a top-level ubuntu_advantage key Pro will automatically attach
# instead of defer to cloud-init for all attach operations.
ubuntu_advantage:
  features:
    disable_auto_attach: true
ubuntu_pro:
  features:
    disable_auto_attach: true
"""

PRO_DAEMON_DISABLED = """\
#cloud-config
# Disable Pro daemon (only needed in GCE)
# Drop ubuntu_advantage key once ubuntu-advantage-tools v. 32 is SRU'd
ubuntu_advantage:
  features:
    disable_auto_attach: true
ubuntu_pro:
  features:
    disable_auto_attach: true
bootcmd:
- sudo systemctl mask ubuntu-advantage.service
"""

AUTO_ATTACH_CUSTOM_SERVICES = """\
#cloud-config
# Drop ubuntu_advantage key once ubuntu-advantage-tools v. 32 is SRU'd
ubuntu_advantage:
  enable:
  - esm-infra
ubuntu_pro:
  enable:
  - esm-infra
"""


def did_ua_service_noop(client: IntegrationInstance) -> bool:
    ua_log = client.read_from_file("/var/log/ubuntu-advantage.log")
    return (
        "Skipping auto-attach and deferring to cloud-init to setup and"
        " configure auto-attach" in ua_log
    )


def is_attached(client: IntegrationInstance) -> bool:
    status_resp = client.execute("sudo pro status --format json")
    assert status_resp.ok
    status = json.loads(status_resp.stdout)
    return bool(status.get("attached"))


def get_services_status(client: IntegrationInstance) -> dict:
    """Creates a map of service -> is_enable.

    pro status --format json contains a key with list of service objects like:

    {
      ...
      "services":[
        {
          "available":"yes",
          "blocked_by":[

          ],
          "description":"Common Criteria EAL2 Provisioning Packages",
          "description_override":null,
          "entitled":"yes",
          "name":"cc-eal",
          "status":"disabled",
          "status_details":"CC EAL2 is not configured"
        },
        ...
      ]
    }

    :return: Dict where the keys are ua service names and the values
    are booleans representing if the service is enable or not.
    """
    status_resp = client.execute("sudo pro status --format json")
    assert status_resp.ok
    status = json.loads(status_resp.stdout)
    return {
        svc["name"]: svc["status"] in ("enabled", "warning")
        for svc in status["services"]
    }


@pytest.mark.adhoc
@pytest.mark.skipif(not IS_UBUNTU, reason="Test is Ubuntu specific")
@pytest.mark.skipif(
    not CLOUD_INIT_UA_TOKEN, reason="CLOUD_INIT_UA_TOKEN env var not provided"
)
class TestUbuntuAdvantage:
    @pytest.mark.user_data(ATTACH_FALLBACK.format(token=CLOUD_INIT_UA_TOKEN))
    def test_valid_token(self, client: IntegrationInstance):
        log = client.read_from_file("/var/log/cloud-init.log")
        verify_clean_log(log)
        assert is_attached(client)
        client.execute("pro detach")
        # Replace ubuntu_pro with previously named ubuntu_advantage
        client.execute(
            "sed -i 's/ubuntu_pro$/ubuntu_advantage/' /etc/cloud/cloud.cfg"
        )
        client.restart()
        status_resp = client.execute("cloud-init status --format json")
        status = json.loads(status_resp.stdout)
        assert (
            "Module has been renamed from cc_ubuntu_advantage to cc_ubuntu_pro"
            in "\n".join(status["recoverable_errors"]["DEPRECATED"])
        )
        assert is_attached(client)

    @pytest.mark.user_data(ATTACH.format(token=CLOUD_INIT_UA_TOKEN))
    def test_idempotency(self, client: IntegrationInstance):
        log = client.read_from_file("/var/log/cloud-init.log")
        verify_clean_log(log)
        assert is_attached(client)

        # Clean reboot to change instance-id and trigger cc_ua in next boot
        assert client.execute("cloud-init clean --logs").ok
        client.restart()

        log = client.read_from_file("/var/log/cloud-init.log")
        verify_clean_log(log)
        assert is_attached(client)

        # Assert service-already-enabled handling for esm-infra.
        # First totally destroy ubuntu-advantage-tools data and state.
        # This is a hack but results in a system that thinks it
        # is detached even though esm-infra is still enabled.
        # When cloud-init runs again, it will successfully re-attach
        # and then notice that esm-infra is already enabled.
        client.execute("rm -rf /var/lib/ubuntu-advantage")
        assert client.execute("cloud-init clean --logs").ok
        client.restart()
        log = client.read_from_file("/var/log/cloud-init.log")
        verify_clean_log(log)
        assert "Service `esm-infra` already enabled" in log


def maybe_install_cloud_init(session_cloud: IntegrationCloud):
    source = get_validated_source(session_cloud)

    launch_kwargs = {
        "image_id": session_cloud.cloud_instance.daily_image(
            CURRENT_RELEASE.series, image_type=ImageType.PRO
        )
    }

    if source is CloudInitSource.NONE:
        LOG.info(
            "No need to customize cloud-init version. Return without spawning"
            " an extra instance"
        )
        return launch_kwargs

    user_data = (
        PRO_DAEMON_DISABLED
        if session_cloud.settings.PLATFORM == "gce"
        else PRO_AUTO_ATTACH_DISABLED
    )

    with session_cloud.launch(
        user_data=user_data,
        launch_kwargs=launch_kwargs,
    ) as client:
        # TODO: Re-enable this check after cloud images contain
        # cloud-init 23.4.
        # Explanation: We have to include something under
        # user-data.ubuntu_pro to skip the automatic auto-attach
        # (driven by ua-auto-attach.service and/or ubuntu-advantage.service)
        # while customizing the instance but in cloud-init < 23.4,
        # user-data.ubuntu_pro requires a token key.

        # log = client.read_from_file("/var/log/cloud-init.log")
        # verify_clean_log(log)

        assert not is_attached(
            client
        ), "Test precondition error. Instance is auto-attached."

        if session_cloud.settings.PLATFORM == "gce":
            LOG.info(
                "Restore `ubuntu-advantage.service` original status for next"
                " boot"
            )
            assert client.execute(
                "sudo systemctl unmask ubuntu-advantage.service"
            ).ok

        client.install_new_cloud_init(source)
        session_cloud.snapshot_id = client.snapshot()
        client.destroy()

    return {"image_id": session_cloud.snapshot_id}


@pytest.mark.skipif(
    not all([IS_UBUNTU, CURRENT_RELEASE in [BIONIC, FOCAL, JAMMY]]),
    reason="Test runs on Ubuntu LTS releases only",
)
@pytest.mark.skipif(
    PLATFORM not in ["azure", "ec2", "gce"],
    reason=f"Pro isn't offered on {PLATFORM}.",
)
class TestUbuntuAdvantagePro:
    def test_custom_services(self, session_cloud: IntegrationCloud):
        launch_kwargs = maybe_install_cloud_init(session_cloud)
        with session_cloud.launch(
            user_data=AUTO_ATTACH_CUSTOM_SERVICES,
            launch_kwargs=launch_kwargs,
        ) as client:
            log = client.read_from_file("/var/log/cloud-init.log")
            verify_clean_log(log)
            assert did_ua_service_noop(client)
            assert is_attached(client)
            services_status = get_services_status(client)
            assert services_status.pop(
                "esm-infra"
            ), "esm-infra expected to be enabled"
            enabled_services = {
                svc for svc, status in services_status.items() if status
            }
            assert (
                not enabled_services
            ), f"Only livepatch must be enabled. Found: {enabled_services}"
